#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os as _os
import sys as _sys

__version__ = '1.9.1'
__version_info__ = tuple(map(int, __version__.split('.')))
__license__ = 'MIT'
__author__ = 'UAVCAN Development Team'

# Our unorthodox approach to dependency management requires us to apply certain workarounds.
# Here, the objective is to allow our library to import stuff from its third-party dependency directory,
# but at the same time we don't want to interfere with the application that depends on this library.
# So we modify the import lookup path temporarily while the package initialization is in progress;
# when done, we restore the path back to its original value. One implication is that it won't be possible
# to import stuff dynamically after the initialization is finished (e.g., function-local imports won't be
# able to reach the third-party stuff), but we don't care.
_original_sys_path = _sys.path
_sys.path = [_os.path.join(_os.path.dirname(__file__), 'third_party')] + _sys.path

# Never import anything that is not available here - API stability guarantees are only provided for the exposed items.
from ._namespace import read_namespace as read_namespace                            # noqa
from ._namespace import PrintOutputHandler as PrintOutputHandler                    # noqa

# Error model.
from ._error import FrontendError as FrontendError                                  # noqa
from ._error import InvalidDefinitionError as InvalidDefinitionError                # noqa
from ._error import InternalError as InternalError                                  # noqa

# Data type model - meta types.
from ._serializable import SerializableType as SerializableType                     # noqa
from ._serializable import PrimitiveType as PrimitiveType                           # noqa
from ._serializable import BooleanType as BooleanType                               # noqa
from ._serializable import ArithmeticType as ArithmeticType                         # noqa
from ._serializable import IntegerType as IntegerType                               # noqa
from ._serializable import SignedIntegerType as SignedIntegerType                   # noqa
from ._serializable import UnsignedIntegerType as UnsignedIntegerType               # noqa
from ._serializable import FloatType as FloatType                                   # noqa
from ._serializable import VoidType as VoidType                                     # noqa
from ._serializable import ArrayType as ArrayType                                   # noqa
from ._serializable import FixedLengthArrayType as FixedLengthArrayType             # noqa
from ._serializable import VariableLengthArrayType as VariableLengthArrayType       # noqa
from ._serializable import CompositeType as CompositeType                           # noqa
from ._serializable import UnionType as UnionType                                   # noqa
from ._serializable import StructureType as StructureType                           # noqa
from ._serializable import DelimitedType as DelimitedType                           # noqa
from ._serializable import ServiceType as ServiceType                               # noqa

# Data type model - attributes.
from ._serializable import Attribute as Attribute                                   # noqa
from ._serializable import Field as Field                                           # noqa
from ._serializable import PaddingField as PaddingField                             # noqa
from ._serializable import Constant as Constant                                     # noqa

# Expression model.
from ._expression import Any as Any                                                 # noqa
from ._expression import Primitive as Primitive                                     # noqa
from ._expression import Boolean as Boolean                                         # noqa
from ._expression import Rational as Rational                                       # noqa
from ._expression import String as String                                           # noqa
from ._expression import Container as Container                                     # noqa
from ._expression import Set as Set                                                 # noqa

# Auxiliary.
from ._serializable import ValueRange as ValueRange                                 # noqa
from ._serializable import Version as Version                                       # noqa
from ._bit_length_set import BitLengthSet as BitLengthSet                           # noqa

_sys.path = _original_sys_path



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
from .. import _expression
from .. import _error
from .._bit_length_set import BitLengthSet


class TypeParameterError(_error.InvalidDefinitionError):
    pass


class SerializableType(_expression.Any):
    """
    Instances are immutable.
    Invoking :meth:`__str__` on a data type returns its uniform normalized definition, e.g.,
    ``uavcan.node.Heartbeat.1.0[<=36]``, ``truncated float16[<=36]``.
    """

    TYPE_NAME = 'metaserializable'

    BITS_PER_BYTE = 8
    """
    This is dictated by the UAVCAN Specification.
    """

    def __init__(self) -> None:
        super(SerializableType, self).__init__()

    @property
    @abc.abstractmethod
    def bit_length_set(self) -> BitLengthSet:
        """
        A set of all possible bit length values of the serialized representations of this type.
        Refer to the specification for the background. The returned set is guaranteed to be non-empty.
        See :class:`pydsdl.BitLengthSet`.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def alignment_requirement(self) -> int:
        """
        Serialized representations of this type are required/guaranteed to be aligned such that their offset
        from the beginning of the containing serialized representation, in bits, is a multiple of this value, in bits.
        Alignment of a type whose alignment requirement is X bits is facilitated by injecting ``[0, X)`` zero
        padding bits before the serialized representation of the type.

        For any element ``L`` of the bit length set of a type whose alignment requirement is ``A``, ``L % A = 0``.
        I.e., the length of a serialized representation of the type is always a multiple of its alignment requirement.

        This value is always a non-negative integer power of two. The alignment of one is a degenerate case denoting
        no alignment.
        """
        raise NotImplementedError

    def _attribute(self, name: _expression.String) -> _expression.Any:
        if name.native_value == '_bit_length_':  # Experimental non-standard extension
            try:
                return _expression.Set(map(_expression.Rational, self.bit_length_set))
            except TypeError:
                pass

        return super(SerializableType, self)._attribute(name)  # Hand over up the inheritance chain, important

    @abc.abstractmethod
    def __str__(self) -> str:   # pragma: no cover
        # Implementations must return a DSDL spec-compatible textual representation of the type.
        # The string representation is used for determining equivalency by the comparison operator __eq__().
        raise NotImplementedError

    def __hash__(self) -> int:
        try:
            bls = self.bit_length_set
        except TypeError:   # If the type is non-serializable.
            bls = BitLengthSet()
        return hash(str(self) + str(bls))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SerializableType):
            same_type = isinstance(other, type(self)) and isinstance(self, type(other))
            try:    # Ensure equality of the bit length sets, otherwise, different types like voids may compare equal.
                same_bls = self.bit_length_set == other.bit_length_set
            except TypeError:   # If the type is non-serializable, assume equality.
                same_bls = same_type
            return same_type and same_bls and str(self) == str(other)
        else:
            return NotImplemented



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import tempfile
from textwrap import dedent
from . import _expression
from . import _error
from . import _parser
from . import _data_type_builder
from . import _dsdl_definition
from . import _serializable
from . import _namespace


# Type annotation disabled here because MyPy is misbehaving, reporting these nonsensical error messages:
#   pydsdl/_test.py:18: error: Missing type parameters for generic type
#   pydsdl/_test.py: note: In function "_in_n_out":
#   pydsdl/_test.py:18: error: Missing type parameters for generic type
_DIRECTORY = None       # type : typing.Optional[tempfile.TemporaryDirectory]


def _parse_definition(definition:         _dsdl_definition.DSDLDefinition,
                      lookup_definitions: typing.Sequence[_dsdl_definition.DSDLDefinition]) \
        -> _serializable.CompositeType:
    return definition.read(lookup_definitions,
                           print_output_handler=lambda line, text: print('Output from line %d:' % line, text),
                           allow_unregulated_fixed_port_id=False)


def _define(rel_path: str, text: str) -> _dsdl_definition.DSDLDefinition:
    rel_path = rel_path.replace('/', os.sep)    # Windows compatibility
    assert _DIRECTORY
    path = os.path.join(_DIRECTORY.name, rel_path)
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w', encoding='utf8') as f:
        f.write(text)

    root_namespace_path = os.path.join(_DIRECTORY.name, rel_path.strip(os.sep).split(os.sep)[0])
    out = _dsdl_definition.DSDLDefinition(path, root_namespace_path)
    print('New definition:', out, 'Root NS:', root_namespace_path)
    return out


def _in_n_out(test: typing.Callable[[], None]) -> typing.Callable[[], None]:
    def decorator() -> None:
        global _DIRECTORY
        _DIRECTORY = tempfile.TemporaryDirectory(prefix='pydsdl-test-')
        try:
            test()
        finally:
            _DIRECTORY = None   # Preserving the contents for future inspection if needed

    return decorator


@_in_n_out
def _unittest_define() -> None:
    # I DON'T ALWAYS WRITE UNIT TESTS
    d = _define('uavcan/test/5000.Message.1.2.uavcan', '# empty')
    assert _DIRECTORY is not None
    assert d.full_name == 'uavcan.test.Message'
    assert d.version == (1, 2)
    assert d.fixed_port_id == 5000
    assert d.file_path == os.path.join(_DIRECTORY.name, 'uavcan', 'test', '5000.Message.1.2.uavcan')
    assert d.root_namespace_path == os.path.join(_DIRECTORY.name, 'uavcan')
    assert open(d.file_path).read() == '# empty'

    # BUT WHEN I DO, I WRITE UNIT TESTS FOR MY UNIT TESTS
    d = _define('uavcan/Service.255.254.uavcan', '# empty 2')
    assert d.full_name == 'uavcan.Service'
    assert d.version == (255, 254)
    assert d.fixed_port_id is None
    assert d.file_path == os.path.join(_DIRECTORY.name, 'uavcan', 'Service.255.254.uavcan')
    assert d.root_namespace_path == os.path.join(_DIRECTORY.name, 'uavcan')
    assert open(d.file_path).read() == '# empty 2'


@_in_n_out
def _unittest_simple() -> None:
    abc = _define(
        'vendor/nested/7000.Abc.1.2.uavcan',
        dedent('''
        @deprecated
        uint8 CHARACTER = '#'
        int8 a
        saturated int64[<33] b
        @extent 1024 * 8
        ''')
    )
    assert abc.fixed_port_id == 7000
    assert abc.full_name == 'vendor.nested.Abc'
    assert abc.version == (1, 2)

    p = _parse_definition(abc, [])
    print('Parsed:', p)
    assert isinstance(p, _serializable.DelimitedType)
    assert isinstance(p.inner_type, _serializable.StructureType)
    assert p.full_name == 'vendor.nested.Abc'
    assert p.source_file_path.endswith(os.path.join('vendor', 'nested', '7000.Abc.1.2.uavcan'))
    assert p.source_file_path == abc.file_path
    assert p.fixed_port_id == 7000
    assert p.deprecated
    assert p.version == (1, 2)
    assert min(p.inner_type.bit_length_set) == 16
    assert max(p.inner_type.bit_length_set) == 16 + 64 * 32
    assert min(p.bit_length_set) == 32
    assert max(p.bit_length_set) == 32 + 1024 * 8
    assert len(p.attributes) == 3
    assert len(p.fields) == 2
    assert str(p.fields[0].data_type) == 'saturated int8'
    assert p.fields[0].name == 'a'
    assert str(p.fields[1].data_type) == 'saturated int64[<=32]'      # Note: normalized representation
    assert p.fields[1].name == 'b'
    assert len(p.constants) == 1
    assert str(p.constants[0].data_type) == 'saturated uint8'
    assert p.constants[0].name == 'CHARACTER'
    assert isinstance(p.constants[0].value, _expression.Rational)
    assert p.constants[0].value == _expression.Rational(ord('#'))

    t = p.fields[1].data_type
    assert isinstance(t, _serializable.ArrayType)
    assert str(t.element_type) == 'saturated int64'

    empty_new = _define(
        'vendor/nested/Empty.255.255.uavcan',
        '''@sealed'''
    )

    empty_old = _define(
        'vendor/nested/Empty.255.254.uavcan',
        '''@sealed'''
    )

    constants = _define(
        'another/Constants.5.0.uavcan',
        dedent('''
        @sealed
        float64 PI = 3.1415926535897932384626433
        ''')
    )

    service = _define(
        'another/300.Service.0.1.uavcan',
        dedent('''
        @union
        @deprecated
        vendor.nested.Empty.255.255 new_empty_implicit
        vendor.nested.Empty.255.255 new_empty_explicit
        vendor.nested.Empty.255.254 old_empty
        @extent 32
        -----------------------------------
        @sealed                      # RESPONSE SEALED REQUEST NOT
        Constants.5.0 constants      # RELATIVE REFERENCE
        vendor.nested.Abc.1.2 abc
        ''')
    )

    p = _parse_definition(service, [
        abc,
        empty_new,
        empty_old,
        constants,
    ])
    print('Parsed:', p)
    assert isinstance(p, _serializable.ServiceType)
    assert p.full_name == 'another.Service'
    assert p.fixed_port_id == 300
    assert p.deprecated
    assert p.version == (0, 1)
    assert not p.constants
    assert p == p
    assert p != empty_new

    assert len(p.fields) == 2
    assert p.fields[0].name == 'request'
    assert p.fields[1].name == 'response'
    req, res = [x.data_type for x in p.fields]
    assert isinstance(req, _serializable.DelimitedType)
    assert isinstance(req.inner_type, _serializable.UnionType)
    assert isinstance(res, _serializable.StructureType)
    assert req.full_name == 'another.Service.Request'
    assert res.full_name == 'another.Service.Response'
    assert req is p.request_type
    assert res is p.response_type
    assert req.has_parent_service
    assert req.inner_type.has_parent_service
    assert res.has_parent_service

    assert len(req.constants) == 0
    assert len(req.fields) == 3
    # noinspection PyUnresolvedReferences
    assert req.inner_type.number_of_variants == 3
    assert req.deprecated
    assert not req.has_fixed_port_id
    assert req.version == (0, 1)
    assert req.bit_length_set == {32, 32 + 8, 32 + 16, 32 + 24, 32 + 32}  # Delimited container, @extent
    assert req.inner_type.bit_length_set == 8                             # Remember this is a union
    assert [x.name for x in req.fields] == ['new_empty_implicit', 'new_empty_explicit', 'old_empty']

    t = req.fields[0].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 255)          # Selected implicitly

    t = req.fields[1].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 255)          # Selected explicitly

    t = req.fields[2].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == 'vendor.nested.Empty'
    assert t.version == (255, 254)          # Selected explicitly

    assert len(res.constants) == 0
    assert len(res.fields) == 2
    assert res.deprecated
    assert not res.has_fixed_port_id
    assert res.version == (0, 1)
    # This is a sealed type, so we get the real BLS, but we mustn't forget about the non-sealed nested field!
    assert min(res.bit_length_set) == 32                                            # Just the delimiter header
    assert max(res.bit_length_set) == 32 + 1024 * 8

    t = res.fields[0].data_type
    assert isinstance(t, _serializable.StructureType)
    assert t.full_name == 'another.Constants'
    assert t.version == (5, 0)

    t = res.fields[1].data_type
    assert isinstance(t, _serializable.DelimitedType)
    assert isinstance(t.inner_type, _serializable.StructureType)
    assert t.full_name == 'vendor.nested.Abc'
    assert t.version == (1, 2)

    p2 = _parse_definition(abc, [
        service,
        empty_new,
        empty_old,
        constants,
    ])
    assert hash(p2) == hash(p2)
    assert hash(p2) != hash(p)
    assert hash(p) == hash(p)

    union = _define(
        'another/Union.5.9.uavcan',
        dedent('''
        @union
        @sealed
        truncated float16 PI = 3.1415926535897932384626433
        uint8 a
        vendor.nested.Empty.255.255[5] b
        saturated bool [ <= 255 ] c
        ''')
    )

    p = _parse_definition(union, [
        empty_old,
        empty_new,
    ])

    assert p.full_name == 'another.Union'
    assert p.version == (5, 9)
    assert p.fixed_port_id is None
    assert not p.has_fixed_port_id
    assert not p.deprecated
    assert isinstance(p, _serializable.UnionType)
    assert p.number_of_variants == 3
    assert len(p.constants) == 1
    assert p.constants[0].name == 'PI'
    assert str(p.constants[0].data_type) == 'truncated float16'
    assert min(p.bit_length_set) == 8
    assert max(p.bit_length_set) == 8 + 8 + 255 + 1  # The last +1 is the padding to byte.
    assert len(p.fields) == 3
    assert str(p.fields[0]) == 'saturated uint8 a'
    assert str(p.fields[1]) == 'vendor.nested.Empty.255.255[5] b'
    assert str(p.fields[2]) == 'saturated bool[<=255] c'


# noinspection PyProtectedMember,PyProtectedMember
@_in_n_out
def _unittest_error() -> None:
    from pytest import raises

    def standalone(rel_path: str, definition: str, allow_unregulated: bool = False) -> _serializable.CompositeType:
        return _define(rel_path, definition + '\n').read([], lambda *_: None, allow_unregulated)  # pragma: no branch

    with raises(_error.InvalidDefinitionError, match='(?i).*port ID.*'):
        standalone('vendor/1000.InvalidRegulatedSubjectID.1.0.uavcan', 'uint2 value\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*port ID.*'):
        standalone('vendor/10.InvalidRegulatedServiceID.1.0.uavcan', 'uint2 v1\n@sealed\n---\nint64 v2\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*subject ID.*'):
        standalone('vendor/100000.InvalidRegulatedSubjectID.1.0.uavcan', 'uint2 value\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*service ID.*'):
        standalone('vendor/1000.InvalidRegulatedServiceID.1.0.uavcan', 'uint2 v1\n@sealed\n---\nint64 v2\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*multiple attributes under the same name.*'):
        standalone('vendor/AttributeNameCollision.1.0.uavcan', 'uint2 value\n@sealed\nint64 value')

    with raises(_error.InvalidDefinitionError, match='(?i).*tagged union cannot contain fewer than.*'):
        standalone('vendor/SmallUnion.1.0.uavcan', '@union\nuint2 value\n@sealed')

    assert standalone('vendor/invalid_constant_value/A.1.0.uavcan',
                      'bool BOOLEAN = false\n@sealed').constants[0].name == 'BOOLEAN'
    with raises(_error.InvalidDefinitionError, match='.*Invalid value for boolean constant.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = 0\n@extent 0')   # Should be false

    with raises(_error.InvalidDefinitionError, match='.*undefined_identifier.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = undefined_identifier\n@extent 0')

    with raises(_parser.DSDLSyntaxError):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'bool BOOLEAN = -\n@extent 0')

    with raises(_error.InvalidDefinitionError, match='(?i).*exceeds the range.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', 'uint10 INTEGRAL = 2000\n@extent 0')

    with raises(_error.InvalidDefinitionError, match='(?i).*character.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "uint8 CH = '\u0451'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match='.*uint8.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "uint9 CH = 'q'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match='.*uint8.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "int8 CH = 'q'\n@extent 0")

    with raises(_error.InvalidDefinitionError, match='.*integer constant.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "int8 CH = 1.1\n@extent 0")

    with raises(_error.InvalidDefinitionError, match='(?i).*type.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "float32 CH = true\n@extent 0")

    with raises(_error.InvalidDefinitionError, match='(?i).*type.*'):
        standalone('vendor/invalid_constant_value/A.1.0.uavcan', "float32 CH = 't'\n@extent 0")

    with raises(_parser.DSDLSyntaxError):
        standalone('vendor/syntax_error/A.1.0.uavcan', 'bool array[10]\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*array capacity.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool[0] array\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*array capacity.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool[<1] array\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*array capacity.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool[true] array\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*array capacity.*'):
        standalone('vendor/array_size/A.1.0.uavcan', 'bool["text"] array\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*service response marker.*'):
        standalone('vendor/service/A.1.0.uavcan',
                   'bool request\n@sealed\n---\nbool response\n@sealed\n---\nbool again\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*unknown directive.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@sho_tse_take\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*requires an expression.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@assert\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*does not expect an expression.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@union true || false\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*does not expect an expression.*'):
        standalone('vendor/directive/A.1.0.uavcan', '@deprecated true || false\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*version number.*'):
        standalone('vendor/version/A.0.0.uavcan', '@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*version number.*'):
        standalone('vendor/version/A.0.256.uavcan', '@sealed')

    with raises(_dsdl_definition.FileNameFormatError):
        standalone('vendor/version/A.0..256.uavcan', '@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*version number.*'):
        standalone('vendor/version/A.256.0.uavcan', '@sealed')

    with raises(_parser.DSDLSyntaxError):
        standalone('vendor/types/A.1.0.uavcan', 'truncated uavcan.node.Heartbeat.1.0 field\n@sealed')

    with raises(_serializable._primitive.InvalidCastModeError):
        standalone('vendor/types/A.1.0.uavcan', 'truncated bool foo\n@sealed')

    with raises(_serializable._primitive.InvalidCastModeError):
        standalone('vendor/types/A.1.0.uavcan', 'truncated int8 foo\n@sealed')

    with raises(_data_type_builder.UndefinedDataTypeError, match=r'(?i).*nonexistent.TypeName.*1\.0.*'):
        standalone('vendor/types/A.1.0.uavcan', 'nonexistent.TypeName.1.0 field\n@sealed')

    with raises(_data_type_builder.UndefinedDataTypeError, match=r"(?i).*vendor[/\\]+types' instead of .*vendor'.*"):
        standalone('vendor/types/A.1.0.uavcan', 'types.Nonexistent.1.0 field\n@sealed')

    with raises(_error.InvalidDefinitionError, match=r'(?i).*not defined for.*'):
        standalone('vendor/types/A.1.0.uavcan',
                   dedent('''
                   @union
                   int8 a
                   @assert _offset_.count >= 1
                   int16 b
                   @sealed
                   '''))

    with raises(_error.InvalidDefinitionError, match=r'(?i).*field offset is not defined for unions.*'):
        standalone('vendor/types/A.1.0.uavcan',
                   dedent('''
                   @union
                   int8 a
                   int16 b
                   @assert _offset_.count >= 1
                   int8 c
                   @sealed
                   '''))

    with raises(_data_type_builder.UndefinedDataTypeError, match=r'.*ns.Type_.*1\.0'):
        _parse_definition(
            _define('vendor/types/A.1.0.uavcan', 'ns.Type_.1.0 field\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
            ]
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*Bit length cannot exceed.*'):
        _parse_definition(
            _define('vendor/types/A.1.0.uavcan', 'int128 field\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
                _define('ns/Type_.1.1.uavcan', '@sealed'),
            ]
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*type.*'):
        _parse_definition(
            _define('vendor/invalid_constant_value/A.1.0.uavcan', 'ns.Type_.1.1 VALUE = 123\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
                _define('ns/Type_.1.1.uavcan', '@sealed'),
            ]
        )

    with raises(_data_type_builder.UndefinedDataTypeError):
        defs = [
            _define('vendor/circular_dependency/A.1.0.uavcan', 'B.1.0 b\n@sealed'),
            _define('vendor/circular_dependency/B.1.0.uavcan', 'A.1.0 b\n@sealed'),
        ]
        _parse_definition(defs[0], defs)

    with raises(_error.InvalidDefinitionError, match='(?i).*union directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan', 'ns.Type_.2.0 field\n@union\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
            ]
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*deprecated directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan', 'ns.Type_.2.0 field\n@deprecated\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
            ]
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*deprecated directive.*'):
        _parse_definition(
            _define('vendor/misplaced_directive/A.1.0.uavcan',
                    'ns.Type_.2.0 field\n@sealed\n---\n@deprecated\n@sealed'),
            [
                _define('ns/Type_.2.0.uavcan', '@sealed'),
            ]
        )

    try:
        standalone('vendor/types/A.1.0.uavcan',
                   dedent('''
                   int8 a  # Comment
                   # Empty
                   @assert false  # Will error here, line number 4
                   # Blank
                   @sealed
                   '''))
    except _error.FrontendError as ex:
        assert ex.path and ex.path.endswith(os.path.join('vendor', 'types', 'A.1.0.uavcan'))
        assert ex.line and ex.line == 4
    else:  # pragma: no cover
        assert False

    standalone('vendor/types/1.A.1.0.uavcan', '@sealed', allow_unregulated=True)
    with raises(_data_type_builder.UnregulatedFixedPortIDError, match=r'.*allow_unregulated_fixed_port_id.*'):
        standalone('vendor/types/1.A.1.0.uavcan', '@sealed')

    standalone('vendor/types/1.A.1.0.uavcan', '@sealed\n---\n@sealed', allow_unregulated=True)
    with raises(_data_type_builder.UnregulatedFixedPortIDError, match=r'.*allow_unregulated_fixed_port_id.*'):
        standalone('vendor/types/1.A.1.0.uavcan', '@sealed\n---\n@sealed')

    with raises(_error.InvalidDefinitionError, match='(?i).*seal.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int8 a
                   @extent 128
                   @sealed
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int8 a
                   @sealed
                   @extent 128
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*sealed.*expression.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int8 a
                   @sealed 12345678
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*expression.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int8 a
                   @extent
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int16 a
                   @extent 8  # Too small
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int16 a
                   @extent {16}  # Wrong type
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*'):
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int16 a
                   @extent 64
                   int8 b
                   '''))

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*'):  # Neither extent nor sealed are specified.
        standalone('vendor/sealing/A.1.0.uavcan',
                   dedent('''
                   int16 a
                   int8 b
                   '''))


@_in_n_out
def _unittest_print() -> None:
    printed_items = None  # type: typing.Optional[typing.Tuple[int, str]]

    def print_handler(line_number: int, text: str) -> None:
        nonlocal printed_items
        printed_items = line_number, text

    _define(
        'ns/A.1.0.uavcan',
        '# line number 1\n'
        '# line number 2\n'
        '@print 2 + 2 == 4   # line number 3\n'
        '# line number 4\n'
        '@sealed\n'
    ).read([], print_handler, False)

    assert printed_items
    assert printed_items[0] == 3
    assert printed_items[1] == 'true'

    _define('ns/B.1.0.uavcan', '@print false\n@sealed').read([], print_handler, False)
    assert printed_items
    assert printed_items[0] == 1
    assert printed_items[1] == 'false'

    _define(
        'ns/Offset.1.0.uavcan',
        '@print _offset_    # Not recorded\n'
        'uint8 a\n'
        '@print _offset_\n'
        '@extent 800\n'
    ).read([], print_handler, False)
    assert printed_items
    assert printed_items[0] == 3
    assert printed_items[1] == "{8}"


# noinspection PyProtectedMember
@_in_n_out
def _unittest_assert() -> None:
    from pytest import raises

    _parse_definition(
        _define(
            'ns/A.1.0.uavcan',
            dedent('''
            @assert _offset_ == {0}
            @assert _offset_.min == _offset_.max
            Array.1.0[2] bar
            @assert _offset_ == {16, 24, 32, 40, 48}
            @assert _offset_.min == 16
            @assert _offset_.max == 48
            @assert _offset_ % 8 == {0}
            @assert _offset_ % 10 == {6, 4, 2, 0, 8}
            @assert _offset_ * 2 == {32, 48, 64, 80, 96}
            @assert 2 * _offset_ == {32, 48, 64, 80, 96}
            @assert _offset_ / 4 == {4, 6, 8, 10, 12}
            @assert _offset_ - 4 == {12, 20, 28, 36, 44}
            @assert _offset_ + 4 == {20, 28, 36, 44, 52}
            uint64 big
            @assert _offset_ - 64 == {16, 24, 32, 40, 48}
            @assert _offset_.min == 80
            @assert _offset_.max == 112
            @assert _offset_.max <= 112
            @assert _offset_.max < 113
            @assert _offset_ == _offset_
            @assert truncated uint64._bit_length_ == {64}
            @assert uint64._bit_length_ == {64}
            @assert Array.1.0._bit_length_.max == 8 + 8 + 8
            @assert Array.1.0._extent_ == 8 + 8 + 8
            @assert Array.1.0._extent_ == Array.1.0._bit_length_.max
            @sealed
            ''')),
        [
            _define('ns/Array.1.0.uavcan', 'uint8[<=2] foo\n@sealed')
        ]
    )

    with raises(_error.InvalidDefinitionError, match='(?i).*operator is not defined.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                dedent('''
                uint64 big
                @assert _offset_ == 64
                @sealed
                ''')),
            []
        )

    with raises(_expression.UndefinedAttributeError):
        _parse_definition(
            _define('ns/C.1.0.uavcan', '@print Service.1.0._bit_length_'),
            [
                _define('ns/Service.1.0.uavcan', 'uint8 a\n@sealed\n---\nuint16 b\n@sealed')
            ]
        )

    with raises(_expression.UndefinedAttributeError):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                '''uint64 LENGTH = uint64.nonexistent_attribute\n@extent 0'''),
            []
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*void.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                'void2 name\n@sealed'),
            []
        )

    with raises(_serializable._attribute.InvalidConstantValueError):
        _parse_definition(_define('ns/C.1.0.uavcan', 'int8 name = true\n@sealed'), [])

    with raises(_error.InvalidDefinitionError, match='.*value.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                'int8 name = {1, 2, 3}\n@sealed'),
            []
        )

    _parse_definition(
        _define(
            'ns/D.1.0.uavcan',
            dedent('''
            @union
            float32 a
            uint64 b
            @assert _offset_ == {40, 72}
            @sealed
            ''')),
        []
    )

    _parse_definition(
        _define(
            'ns/E.1.0.uavcan',
            dedent('''
            @union
            uint8 A = 0
            float32 a
            uint8 B = 1
            uint64 b
            uint8 C = 2
            @assert _offset_ == {40, 72}
            uint8 D = 3
            @sealed
            ''')),
        []
    )

    with raises(_error.InvalidDefinitionError):
        _parse_definition(
            _define(
                'ns/F.1.0.uavcan',
                dedent('''
                @union
                @assert _offset_.min == 33
                float32 a
                uint64 b
                @assert _offset_ == {40, 72}
                @sealed
                ''')),
            []
        )

    with raises(_data_type_builder.AssertionCheckFailureError):
        _parse_definition(
            _define(
                'ns/G.1.0.uavcan',
                dedent('''
                float32 a
                @assert _offset_.min == 8
                @sealed
                ''')),
            []
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*yield a boolean.*'):
        _parse_definition(
            _define(
                'ns/H.1.0.uavcan',
                dedent('''
                float32 a
                @assert _offset_.min
                @sealed
                ''')),
            []
        )

    # Extent verification
    _parse_definition(
        _define(
            'ns/I.1.0.uavcan',
            dedent('''
            @assert J.1.0._extent_ == 64
            @assert J.1.0._bit_length_ == {0, 1, 2, 3, 4, 5, 6, 7, 8} * 8 + 32
            @assert K.1.0._extent_ == 8
            @assert K.1.0._bit_length_ == {8}
            @sealed
            ''')
        ),
        [
            _define('ns/J.1.0.uavcan', 'uint8 foo\n@extent 64'),
            _define('ns/K.1.0.uavcan', 'uint8 foo\n@sealed'),
        ]
    )

    # Alignment
    _parse_definition(
        _define(
            'ns/L.1.0.uavcan',
            dedent('''
            @assert _offset_ == {0}
            uint3 a
            @assert _offset_ == {3}
            N.1.0 nothing
            @print _offset_
            @assert _offset_ == {8}   # Aligned!
            uint5 b
            @assert _offset_ == {13}
            N.1.0[3] array_of_nothing
            @assert _offset_ == {16}  # Aligned!
            bool c
            @assert _offset_ == {17}
            M.1.0 variable
            @assert _offset_ == 32 + {24, 32, 40}  # Aligned; variability due to extensibility (non-sealing)
            @sealed
            ''')
        ),
        [
            _define('ns/M.1.0.uavcan', '@extent 16'),
            _define('ns/N.1.0.uavcan', '@sealed'),
        ]
    )


def _unittest_parse_namespace() -> None:
    from pytest import raises
    directory = tempfile.TemporaryDirectory()

    print_output = None  # type: typing.Optional[typing.Tuple[str, int, str]]

    def print_handler(d: str, line: int, text: str) -> None:
        nonlocal print_output
        print_output = d, line, text

    # noinspection PyShadowingNames
    def _define(rel_path: str, text: str) -> None:
        path = os.path.join(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write(text)

    # Empty namespace.
    assert [] == _namespace.read_namespace(directory.name)

    _define(
        'zubax/First.1.0.uavcan',
        dedent("""
        uint8[<256] a
        @assert _offset_.min == 8
        @assert _offset_.max == 2048
        @sealed
        """)
    )

    _define(
        'zubax/7001.Message.1.0.uavcan',
        dedent("""
        zubax.First.1.0[<=2] a
        @assert _offset_.min == 8
        @assert _offset_.max == 4104
        @extent _offset_.max * 8
        """)
    )

    _define(
        'zubax/nested/300.Spartans.30.0.uavcan',
        dedent("""
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        @extent _offset_.max * 8
        ---
        @print _offset_     # Will print zero {0}
        @sealed
        """)
    )

    _define('zubax/nested/300.Spartans.30.0.txt', 'completely unrelated stuff')
    _define('zubax/300.Spartans.30.0', 'completely unrelated stuff')

    parsed = _namespace.read_namespace(
        os.path.join(directory.name, 'zubax'),
        [],
        print_handler
    )
    print(parsed)
    assert len(parsed) == 3
    assert 'zubax.First' in [x.full_name for x in parsed]
    assert 'zubax.Message' in [x.full_name for x in parsed]
    assert 'zubax.nested.Spartans' in [x.full_name for x in parsed]

    # try again with minimal arguments to read_namespace
    parsed_minimal_args = _namespace.read_namespace(
        os.path.join(directory.name, 'zubax')
    )
    assert len(parsed_minimal_args) == 3

    _define(
        'zubax/colliding/300.Iceberg.30.0.uavcan',
        dedent("""
        @extent 1024
        ---
        @extent 1024
        """)
    )

    with raises(_namespace.FixedPortIDCollisionError):
        _namespace.read_namespace(
            os.path.join(directory.name, 'zubax'),
            [],
            print_handler
        )

    with raises(TypeError):  # Invalid usage: expected path-like object, not bytes.
        _namespace.read_namespace(os.path.join(directory.name, 'zubax'), b'/my/path')  # type: ignore

    with raises(TypeError):  # Invalid usage: expected path-like object, not bytes.
        # noinspection PyTypeChecker
        _namespace.read_namespace(os.path.join(directory.name, 'zubax'), [b'/my/path'])  # type: ignore

    assert print_output is not None
    assert '300.Spartans' in print_output[0]
    assert print_output[1] == 9
    assert print_output[2] == '{0}'

    _define(
        'zubax/colliding/iceberg/300.Ice.30.0.uavcan',
        dedent("""
        @sealed
        ---
        @sealed
        """)
    )
    with raises(_namespace.DataTypeNameCollisionError):
        _namespace.read_namespace(os.path.join(directory.name, 'zubax'), [
            os.path.join(directory.name, 'zubax'),
        ])

    # Do again to test single lookup-directory override
    with raises(_namespace.DataTypeNameCollisionError):
        _namespace.read_namespace(os.path.join(directory.name, 'zubax'),
                                  os.path.join(directory.name, 'zubax'))

    try:
        os.unlink(os.path.join(directory.name, 'zubax/colliding/iceberg/300.Ice.30.0.uavcan'))
        _define(
            'zubax/COLLIDING/300.Iceberg.30.0.uavcan',
            dedent("""
            @extent 1024
            ---
            @extent 1024
            """)
        )
        with raises(_namespace.DataTypeNameCollisionError, match='.*letter case.*'):
            _namespace.read_namespace(os.path.join(directory.name, 'zubax'), [
                os.path.join(directory.name, 'zubax'),
            ])
    except _namespace.FixedPortIDCollisionError:  # pragma: no cover
        pass  # We're running on a platform where paths are not case-sensitive.


def _unittest_parse_namespace_versioning() -> None:
    from pytest import raises
    import glob
    directory = tempfile.TemporaryDirectory()

    # noinspection PyShadowingNames
    def _define(rel_path: str, text: str) -> None:
        path = os.path.join(directory.name, rel_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as f:
            f.write(text)

    def _undefine_glob(rel_path_glob: str) -> None:
        path = os.path.join(directory.name, rel_path_glob)
        for g in glob.glob(path):
            os.remove(g)

    _define(
        'ns/Spartans.30.0.uavcan',
        dedent("""
        @deprecated
        @union
        float16 small
        float32 just_right
        float64 woah
        @extent 1024
        ---
        @extent 1024
        """)
    )

    _define(
        'ns/Spartans.30.1.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        ---
        @extent 1024
        """)
    )

    parsed = _namespace.read_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    print(parsed)
    assert len(parsed) == 2

    _define(
        'ns/Spartans.30.2.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    with raises(_namespace.VersionsOfDifferentKindError):
        _namespace.read_namespace(
            os.path.join(directory.name, 'ns'),
            []
        )

    _undefine_glob('ns/Spartans.30.[01].uavcan')

    _define(
        'ns/Spartans.30.0.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    parsed = _namespace.read_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    print(parsed)
    assert len(parsed) == 2

    _define(
        'ns/Spartans.30.1.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        int64 woah
        @extent 1024
        """)
    )

    _define(
        'ns/6700.Spartans.30.2.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        int32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    with raises(_namespace.MultipleDefinitionsUnderSameVersionError):
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])

    _undefine_glob('ns/Spartans.30.2.uavcan')

    parsed = _namespace.read_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    assert len(parsed) == 3

    _undefine_glob('ns/Spartans.30.0.uavcan')
    _define(
        'ns/6700.Spartans.30.0.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    with raises(_namespace.MinorVersionFixedPortIDError):
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])

    _undefine_glob('ns/Spartans.30.1.uavcan')
    _define(
        'ns/6700.Spartans.30.1.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    parsed = _namespace.read_namespace(
        os.path.join(directory.name, 'ns'),
        []
    )
    assert len(parsed) == 3

    _undefine_glob('ns/6700.Spartans.30.1.uavcan')
    _define(
        'ns/6701.Spartans.30.1.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    with raises(_namespace.MinorVersionFixedPortIDError):
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])

    # Adding new major version under the same FPID
    _undefine_glob('ns/6701.Spartans.30.1.uavcan')
    _define(
        'ns/6700.Spartans.31.0.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    with raises(_namespace.FixedPortIDCollisionError):
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])

    # Major version zero allows us to re-use the same FPID under a different (non-zero) major version
    _undefine_glob('ns/6700.Spartans.31.0.uavcan')
    _define(
        'ns/6700.Spartans.0.1.uavcan',
        dedent("""
        @deprecated
        @union
        uint16 small
        float32 just_right
        float64[1] woah
        @extent 1024
        """)
    )

    # These are needed to ensure full branch coverage, see the checking code.
    _define('ns/Empty.1.0.uavcan', '@extent 0')
    _define('ns/Empty.1.1.uavcan', '@extent 0')
    _define('ns/Empty.2.0.uavcan', '@extent 0')
    _define('ns/6800.Empty.3.0.uavcan', '@extent 0')
    _define('ns/6801.Empty.4.0.uavcan', '@extent 0')

    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 8

    # Check ordering - the definitions must be sorted properly by name (lexicographically) and version (newest first).
    assert list(map(str, parsed)) == [
        'ns.Empty.4.0',         # Newest first
        'ns.Empty.3.0',
        'ns.Empty.2.0',
        'ns.Empty.1.1',
        'ns.Empty.1.0',
        'ns.Spartans.30.2',     # Newest first
        'ns.Spartans.30.0',
        'ns.Spartans.0.1',
    ]

    # Extent consistency -- non-service type
    _define('ns/Consistency.1.0.uavcan', 'uint8 a\n@extent 128')
    _define('ns/Consistency.1.1.uavcan', 'uint8 a\nuint8 b\n@extent 128')
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan', 'uint8 a\nuint8 b\nuint8 c\n@extent 256')
    with raises(_namespace.ExtentConsistencyError,
                match=r'(?i).*extent of ns\.Consistency\.1\.2 is 256 bits.*') as ei_extent:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_extent.value)
    assert ei_extent.value.path and 'Consistency.1' in ei_extent.value.path
    _undefine_glob('ns/Consistency*')

    # Extent consistency -- request
    _define('ns/Consistency.1.0.uavcan',
            dedent('''
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 128
            '''))
    _define('ns/Consistency.1.1.uavcan',
            dedent('''
            uint8 a
            uint8 b
            @extent 128
            ---
            uint8 a
            @extent 128
            '''))
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan',
            dedent('''
            uint8 a
            uint8 b
            @extent 256
            ---
            uint8 a
            @extent 128
            '''))
    with raises(_namespace.ExtentConsistencyError,
                match=r'(?i).*extent of ns\.Consistency.* is 256 bits.*') as ei_extent:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_extent.value)
    assert ei_extent.value.path and 'Consistency.1' in ei_extent.value.path
    _undefine_glob('ns/Consistency*')

    # Extent consistency -- response
    _define('ns/Consistency.1.0.uavcan',
            dedent('''
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 128
            '''))
    _define('ns/Consistency.1.1.uavcan',
            dedent('''
            uint8 a
            @extent 128
            ---
            uint8 a
            uint8 b
            @extent 128
            '''))
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan',
            dedent('''
            uint8 a
            @extent 128
            ---
            uint8 a
            @extent 256
            '''))
    with raises(_namespace.ExtentConsistencyError,
                match=r'(?i).*extent of ns\.Consistency.* is 256 bits.*') as ei_extent:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_extent.value)
    assert ei_extent.value.path and 'Consistency.1' in ei_extent.value.path
    _undefine_glob('ns/Consistency*')

    # Sealing consistency -- non-service type
    _define('ns/Consistency.1.0.uavcan', 'uint64 a\n@extent 64')
    _define('ns/Consistency.1.1.uavcan', 'uint64 a\n@extent 64')
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan', 'uint64 a\n@sealed')
    with raises(_namespace.SealingConsistencyError, match=r'(?i).*ns\.Consistency\.1\.2 is sealed.*') as ei_sealing:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and 'Consistency.1' in ei_sealing.value.path
    _undefine_glob('ns/Consistency*')

    # Sealing consistency -- request
    _define('ns/Consistency.1.0.uavcan',
            dedent('''
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            '''))
    _define('ns/Consistency.1.1.uavcan',
            dedent('''
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            '''))
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan',
            dedent('''
            uint64 a
            @sealed
            ---
            uint64 a
            @extent 64
            '''))
    with raises(_namespace.SealingConsistencyError, match=r'(?i).*ns\.Consistency.* is sealed.*') as ei_sealing:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and 'Consistency.1' in ei_sealing.value.path
    _undefine_glob('ns/Consistency*')

    # Sealing consistency -- response
    _define('ns/Consistency.1.0.uavcan',
            dedent('''
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            '''))
    _define('ns/Consistency.1.1.uavcan',
            dedent('''
            uint64 a
            @extent 64
            ---
            uint64 a
            @extent 64
            '''))
    parsed = _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])     # no error
    assert len(parsed) == 10
    _define('ns/Consistency.1.2.uavcan',
            dedent('''
            uint64 a
            @extent 64
            ---
            uint64 a
            @sealed
            '''))
    with raises(_namespace.SealingConsistencyError, match=r'(?i).*ns\.Consistency.* is sealed.*') as ei_sealing:
        _namespace.read_namespace(os.path.join(directory.name, 'ns'), [])
    print(ei_sealing.value)
    assert ei_sealing.value.path and 'Consistency.1' in ei_sealing.value.path
    _undefine_glob('ns/Consistency*')


def _unittest_parse_namespace_faults() -> None:
    try:
        _namespace.read_namespace('/foo/bar/baz', ['/bat/wot', '/foo/bar/baz/bad'])
    except _namespace.NestedRootNamespaceError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False

    try:
        _namespace.read_namespace('/foo/bar/baz', ['/foo/bar/zoo', '/foo/bar/doo/roo/BAZ'])  # Notice the letter case
    except _namespace.RootNamespaceNameCollisionError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False
    try:
        _namespace.read_namespace('/foo/bar/baz', ['/foo/bar/zoo', '/foo/bar/doo/roo/zoo', '/foo/bar/doo/roo/baz'])
    except _namespace.RootNamespaceNameCollisionError as ex:
        print(ex)
    else:               # pragma: no cover
        assert False


@_in_n_out
def _unittest_inconsistent_deprecation() -> None:
    from pytest import raises

    _parse_definition(
        _define('ns/A.1.0.uavcan', '@sealed'),
        [
            _define('ns/B.1.0.uavcan',
                    dedent('''
                    @deprecated
                    A.1.0 a
                    @sealed
                    '''))
        ]
    )

    with raises(_error.InvalidDefinitionError, match='(?i).*depend.*deprecated.*'):
        _parse_definition(
            _define(
                'ns/C.1.0.uavcan',
                dedent('''
                X.1.0 b
                @sealed
                ''')
            ),
            [
                _define('ns/X.1.0.uavcan', '@deprecated\n@sealed')
            ]
        )

    _parse_definition(
        _define('ns/D.1.0.uavcan',
                dedent('''
                @deprecated
                X.1.0 b
                @sealed
                ''')),
        [
            _define('ns/X.1.0.uavcan', '@deprecated\n@sealed')
        ]
    )


@_in_n_out
def _unittest_repeated_directives() -> None:
    from pytest import raises

    _parse_definition(
        _define('ns/A.1.0.uavcan',
                dedent('''
                @union
                @deprecated
                int8 a
                float16 b
                @sealed
                ''')),
        []
    )

    with raises(_error.InvalidDefinitionError, match='(?i).*deprecated.*'):
        _parse_definition(
            _define('ns/A.1.0.uavcan',
                    dedent('''
                    @deprecated
                    @deprecated
                    @sealed
                    ''')),
            []
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*deprecated.*'):
        _parse_definition(
            _define('ns/A.1.0.uavcan',
                    dedent('''
                    @deprecated
                    @sealed
                    ---
                    @deprecated
                    @sealed
                    ''')),
            []
        )

    _parse_definition(
        _define('ns/A.1.0.uavcan',
                dedent('''
                @union
                int8 a
                float16 b
                @sealed
                ---
                @union
                int8 a
                float16 b
                @sealed
                ''')),
        []
    )

    with raises(_error.InvalidDefinitionError, match='(?i).*union.*'):
        _parse_definition(
            _define('ns/A.1.0.uavcan',
                    dedent('''
                    @union
                    @union
                    int8 a
                    float16 b
                    @sealed
                    ''')),
            []
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*sealed.*'):
        _parse_definition(
            _define('ns/A.1.0.uavcan',
                    dedent('''
                    @sealed
                    @sealed
                    int8 a
                    float16 b
                    @sealed
                    ''')),
            []
        )

    with raises(_error.InvalidDefinitionError, match='(?i).*extent.*already set.*'):
        _parse_definition(
            _define('ns/A.1.0.uavcan',
                    dedent('''
                    int8 a
                    float16 b
                    @extent 256
                    @extent 800
                    @sealed
                    ''')),
            []
        )


@_in_n_out
def _unittest_dsdl_parser_basics() -> None:
    # This is how you can run one test only for development needs:
    #   pytest pydsdl -k _unittest_dsdl_parser_basics --capture=no
    # noinspection SpellCheckingInspection
    _parse_definition(
        _define('ns/A.1.0.uavcan',
                dedent(r'''
                @deprecated
                void16
                int8           [<=123+456] array_inclusive
                truncated uint8[< 123+456] array_exclusive
                saturated int8 [  123+456] array_fixed
                #ns.Bar.1.23 field
                float64 a = +10 * (-2 / -3) / 4 % 5
                bool    b = !true
                float32 c = (123456 + 0x_ab_cd_ef) / 0b1111_1111 ** 2 - 0o123_456 * 2.7
                @print "Hello\r\nworld!"
                @print
                @assert true
                @assert ns.Foo.1.0.THE_CONSTANT == 42
                @assert ns.Bar.1.23.B == ns.Bar.1.23.A + 1
                @extent 32 * 1024 * 8
                ''')),
        [
            _define('ns/Foo.1.0.uavcan', 'int8 THE_CONSTANT = 42\n@extent 1024'),
            _define('ns/Bar.1.23.uavcan', 'int8 the_field\nint8 A = 0xA\nint8 B = 0xB\n@extent 1024'),
        ]
    )


@_in_n_out
def _unittest_dsdl_parser_expressions() -> None:
    from pytest import raises

    def throws(definition: str, exc: typing.Type[Exception] = _expression.InvalidOperandError) -> None:
        with raises(exc):
            _parse_definition(_define('ns/Throws.0.1.uavcan', dedent(definition + '\n@sealed')), [])

    throws('bool R = true && 0')
    throws('bool R = true || 0')
    throws('bool R = 0 || true')
    throws('bool R = 0 == true')
    throws('bool R = {0} & true')
    throws('bool R = true ^ {0}')
    throws('bool R = 0 ^ true')
    throws('int8 R = 1 / 0')
    throws('bool R = "S" == 0')
    throws('bool R = {0} != {}')
    throws('bool R = {0, true, "S"}')
    throws('bool R = {0} == {"s"}')
    throws('bool R = {0} <= "s"')
    throws('bool R = {0} >= "s"')
    throws('bool R = {0} > "s"')
    throws('bool R = {0} < "s"')
    throws('bool R = {0} | "s"')
    throws('bool R = {0} & "s"')
    throws('bool R = {0} ^ "s"')
    throws('bool R = {0}.nonexistent_attribute')
    throws('bool R = {0} / {1}')
    throws('bool R = !1')
    throws('bool R = +true')
    throws('bool R = -"1"')
    throws('bool R = true | false')
    throws('bool R = true & false')
    throws('bool R = true + "0"')
    throws('bool R = true - "0"')
    throws('bool R = true * "0"')
    throws('bool R = true / "0"')
    throws('bool R = true % "0"')
    throws('bool R = true ** "0"')

    _parse_definition(
        _define('ns/A.1.0.uavcan',
                dedent(r'''
                float64 PI = 3.141592653589793
                float64 E  = 2.718281828459045
                @assert (PI ** E > 22.4) && (PI ** E < 22.5)
                @assert 'moments of eternity'     != "strangers stealing someone else's dreams"  # I've seen it all
                @assert 'hunting for the mystery' != 'running for your life in times like these' # I've seen it all
                @assert "I remember the time once in a life" != 'oh baby'  # got you here in my head, here in my head
                @assert false == ('oh' == 'maybe')
                @assert true
                @assert 1 == 2 - 1
                @assert -10 == +20 / -2
                @assert {10, 15, 20} % 5 == {0}
                @assert {10, 15, 20} % 5 == {1, 2, 3} * 0
                @assert {10, 15, 20} / 5 == {2, 3, 4} * 10 / 5 / 2
                @assert {1} < {1, 2}
                @assert {1} <= {1}
                @assert {1} != {1, 2}
                @assert {1} >= {1}
                @assert {1, 2} > {1, 2} == false
                @assert {1, 2, 3} > {1, 2}
                @assert {1, 5/2} == {-5/-2, 2.5, 1}
                @assert {1, 2, 3} == {1} | {2} | {3} | {1}
                @assert {1, 2, 3} == {1, 2, 3, 4, 5} & {1, 2, 3, 8, 9}
                @assert {4, 5, 8, 9} == {1, 2, 3, 4, 5} ^ {1, 2, 3, 8, 9}
                @assert 1 - {1, 2, 3} == {0, -1, -2}
                @assert 1 / {1, 2, 3} == {1, 1/2, 1/3}
                @assert 8 % {1, 2, 3} == {0, 2}
                @assert 2 ** {1, 2, 3} == {2, 4, 8}
                @assert {1, 2, 3} ** 2 == {1, 4, 9}
                @assert "Hello" + ' ' + 'world' == 'Hello world'
                @assert 'Hello'+' '+'world' != ''
                @assert '\u00e9' == '\u0065\u0301'  # e with accent
                @assert '\ufb03' != 'ffi'           # ffi ligature is not decomposed by NFC
                @assert true != ('A' == "a")
                @assert true && true
                @assert ! (true && false)
                @assert true||false
                @assert !false
                @assert ! 5 < 3
                @assert 4 | 2 == 6
                @assert 4 & 2 == 0
                @assert 3 & 2 == 2
                @assert 0xFF_00 & 0x00_FF == 0x0000
                @assert 0xFF_00 | 0x00_FF == 0xFFFF
                @assert 0xFF_00 ^ 0x0F_FF == 0xF0FF
                @sealed
                ''')),
        []
    )


@_in_n_out
def _unittest_pickle() -> None:
    import pickle
    p = _parse_definition(
        _define('ns/A.1.0.uavcan',
                dedent(r'''
                float64 PI = 3.141592653589793
                float64 big_pi
                @sealed
                ---
                float16 small_pi
                @extent 1024 * 8
                ''')),
        []
    )
    assert isinstance(p, _serializable.ServiceType)
    assert p.request_type.has_parent_service
    assert p.response_type.has_parent_service
    assert not p.has_parent_service

    pp = pickle.loads(pickle.dumps(p))
    assert isinstance(pp, _serializable.ServiceType)
    assert pp.request_type.has_parent_service
    assert pp.response_type.has_parent_service
    assert not pp.has_parent_service
    assert str(pp) == str(p)
    assert repr(pp) == repr(p)


def _collect_descendants(cls: typing.Type[object]) -> typing.Iterable[typing.Type[object]]:
    # noinspection PyArgumentList
    for t in cls.__subclasses__():
        yield t
        yield from _collect_descendants(t)


def _unittest_collect_descendants() -> None:  # Unit test for my unit test.
    class A:
        pass

    class B(A):
        pass

    class C(B):
        pass

    class D(A):
        pass

    assert set(_collect_descendants(A)) == {B, C, D}
    assert set(_collect_descendants(D)) == set()
    assert bool in set(_collect_descendants(int))


def _unittest_public_api() -> None:
    import pydsdl

    # Ensure that all descendants of the specified classes are exported from the library.
    # If this test fails, you probably forgot to update __init__.py.
    public_roots = [
        _serializable.SerializableType,
        _serializable.Attribute,
        _expression.Any,
    ]

    for root in public_roots:
        expected_types = {root} | set(_collect_descendants(root))
        for t in expected_types:
            assert t.__name__ in dir(pydsdl), 'Data type %r is not exported' % t



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import re
import string
import typing  # pylint: disable=W0611
from ._serializable import TypeParameterError


class InvalidNameError(TypeParameterError):
    pass


def check_name(name: str) -> None:
    """
    Ensure that the name complies with the requirements set out in the Specification;
    raise :class:`InvalidNameError` if not.
    """
    if not name:
        raise InvalidNameError('Name or namespace component cannot be empty')

    name = name.lower()

    if name[0] not in _VALID_FIRST_CHARACTERS_OF_NAME:
        raise InvalidNameError('Name or namespace component cannot start with %r' % name[0])

    for char in name:
        if char not in _VALID_CONTINUATION_CHARACTERS_OF_NAME:
            raise InvalidNameError('Name or namespace component cannot contain %r' % char)

    for pat in _DISALLOWED_NAME_PATTERNS:
        if isinstance(pat, str):
            if pat == name:
                raise InvalidNameError('Disallowed name: %r matches the following string: %s' % (name, pat))
        elif pat.match(name):
            raise InvalidNameError('Disallowed name: %r matches the following pattern: %s' % (name, pat))


_VALID_FIRST_CHARACTERS_OF_NAME = string.ascii_letters + '_'
_VALID_CONTINUATION_CHARACTERS_OF_NAME = _VALID_FIRST_CHARACTERS_OF_NAME + string.digits

# Disallowed name patterns apply to any part of any name, e.g., an attribute name, a namespace component,
# type name, etc. The pattern must produce an exact match to trigger a name error. All patterns are case-insensitive.
_DISALLOWED_NAME_PATTERNS = [
    'truncated',
    'saturated',
    'true',
    'false',
    'bool',
    re.compile(r'void\d*$'),
    re.compile(r'u?int\d*$'),
    re.compile(r'u?q\d+_\d+$'),
    re.compile(r'float\d*$'),
    'optional',
    'aligned',
    'const',
    'struct',
    'super',
    'template',
    'enum',
    'self',
    'and',
    'or',
    'not',
    'auto',
    'type',
    'con',
    'prn',
    'aux',
    'nul',
    re.compile(r'com\d$'),
    re.compile(r'lpt\d$'),
    re.compile(r'_.*_$')
]  # type: typing.List[typing.Union[str, typing.Pattern[str]]]


def _unittest_check_name() -> None:
    from pytest import raises

    check_name('abc')
    check_name('_abc')
    check_name('abc_')
    check_name('abc0')

    with raises(InvalidNameError):
        check_name('0abc')

    with raises(InvalidNameError):
        check_name('_abc_')

    with raises(InvalidNameError):
        check_name('a-bc')

    with raises(InvalidNameError):
        check_name('')

    with raises(InvalidNameError):
        check_name('truncated')

    with raises(InvalidNameError):
        check_name('COM1')

    with raises(InvalidNameError):
        check_name('Aux')

    with raises(InvalidNameError):
        check_name('float128')

    with raises(InvalidNameError):
        check_name('q16_8')

    with raises(InvalidNameError):
        check_name('uq1_32')



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
import enum
import math
import typing
import fractions
from .._bit_length_set import BitLengthSet
from ._serializable import SerializableType, TypeParameterError


ValueRange = typing.NamedTuple('ValueRange', [('min', fractions.Fraction), ('max', fractions.Fraction)])


class InvalidBitLengthError(TypeParameterError):
    pass


class InvalidCastModeError(TypeParameterError):
    pass


class PrimitiveType(SerializableType):
    MAX_BIT_LENGTH = 64
    BITS_IN_BYTE = 8  # Defined in the UAVCAN specification

    class CastMode(enum.Enum):
        SATURATED = 0
        TRUNCATED = 1

    def __init__(self,
                 bit_length: int,
                 cast_mode: 'PrimitiveType.CastMode'):
        super(PrimitiveType, self).__init__()
        self._bit_length = int(bit_length)
        self._cast_mode = cast_mode

        if self._bit_length < 1:
            raise InvalidBitLengthError('Bit length must be positive')

        if self._bit_length > self.MAX_BIT_LENGTH:
            raise InvalidBitLengthError('Bit length cannot exceed %r' % self.MAX_BIT_LENGTH)

        self._standard_bit_length = \
            (self._bit_length >= self.BITS_IN_BYTE) and (2 ** round(math.log2(self._bit_length)) == self._bit_length)

    @property
    def bit_length_set(self) -> BitLengthSet:
        return BitLengthSet(self.bit_length)

    @property
    def bit_length(self) -> int:
        """
        This is a shortcut for ``next(iter(x.bit_length_set))``, because the bit length set of a primitive type
        always contains exactly one element (i.e., primitive types are fixed-length).
        """
        return self._bit_length

    @property
    def standard_bit_length(self) -> bool:
        """
        The term "standard length" here means that values of such bit length are commonly used in modern computer
        microarchitectures, such as ``uint8``, ``float64``, ``int32``, and so on. Booleans are excluded.
        More precisely, a primitive is said to be "standard length" when the following hold::

            bit_length >= 8
            2**ceil(log2(bit_length)) == bit_length.
        """
        return self._standard_bit_length

    @property
    def cast_mode(self) -> 'PrimitiveType.CastMode':
        return self._cast_mode

    @property
    def alignment_requirement(self) -> int:
        return 1

    @property
    def _cast_mode_name(self) -> str:
        """For internal use only."""
        return {
            self.CastMode.SATURATED: 'saturated',
            self.CastMode.TRUNCATED: 'truncated',
        }[self.cast_mode]

    @abc.abstractmethod
    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError

    def __repr__(self) -> str:
        return '%s(bit_length=%r, cast_mode=%r)' % (self.__class__.__name__, self.bit_length, self.cast_mode)


class BooleanType(PrimitiveType):
    def __init__(self, cast_mode: PrimitiveType.CastMode):
        super(BooleanType, self).__init__(bit_length=1, cast_mode=cast_mode)

        if cast_mode != PrimitiveType.CastMode.SATURATED:
            raise InvalidCastModeError('Invalid cast mode for boolean: %r' % cast_mode)

    def __str__(self) -> str:
        return self._cast_mode_name + ' bool'


class ArithmeticType(PrimitiveType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(ArithmeticType, self).__init__(bit_length, cast_mode)

    @property
    @abc.abstractmethod
    def inclusive_value_range(self) -> ValueRange:   # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class IntegerType(ArithmeticType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(IntegerType, self).__init__(bit_length, cast_mode)

    @property
    @abc.abstractmethod
    def inclusive_value_range(self) -> ValueRange:   # pragma: no cover
        raise NotImplementedError

    @abc.abstractmethod
    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class SignedIntegerType(IntegerType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(SignedIntegerType, self).__init__(bit_length, cast_mode)

        if self._bit_length < 2:
            raise InvalidBitLengthError('Bit length of signed integer types cannot be less than 2')

        if cast_mode != PrimitiveType.CastMode.SATURATED:
            raise InvalidCastModeError('Invalid cast mode for signed integer: %r' % cast_mode)

    @property
    def inclusive_value_range(self) -> ValueRange:
        uint_max_half = ((1 << self.bit_length) - 1) // 2
        return ValueRange(min=fractions.Fraction(-uint_max_half - 1),
                          max=fractions.Fraction(+uint_max_half))

    def __str__(self) -> str:
        return self._cast_mode_name + ' int' + str(self.bit_length)


class UnsignedIntegerType(IntegerType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(UnsignedIntegerType, self).__init__(bit_length, cast_mode)

    @property
    def inclusive_value_range(self) -> ValueRange:
        return ValueRange(min=fractions.Fraction(0),
                          max=fractions.Fraction((1 << self.bit_length) - 1))

    def __str__(self) -> str:
        return self._cast_mode_name + ' uint' + str(self.bit_length)


class FloatType(ArithmeticType):
    def __init__(self,
                 bit_length: int,
                 cast_mode: PrimitiveType.CastMode):
        super(FloatType, self).__init__(bit_length, cast_mode)

        try:
            frac = fractions.Fraction
            # The limits are exact
            self._magnitude = fractions.Fraction({
                16: (2 ** 0x00F) * (2 - frac(2) ** frac(-10)),   # IEEE 754 binary16
                32: (2 ** 0x07F) * (2 - frac(2) ** frac(-23)),   # IEEE 754 binary32
                64: (2 ** 0x3FF) * (2 - frac(2) ** frac(-52)),   # IEEE 754 binary64
            }[self.bit_length])
        except KeyError:
            raise InvalidBitLengthError('Invalid bit length for float type: %d' % bit_length) from None

    @property
    def inclusive_value_range(self) -> ValueRange:
        return ValueRange(min=-self._magnitude,
                          max=+self._magnitude)

    def __str__(self) -> str:
        return self._cast_mode_name + ' float' + str(self.bit_length)


def _unittest_primitive() -> None:
    from pytest import raises, approx

    assert str(BooleanType(PrimitiveType.CastMode.SATURATED)) == 'saturated bool'

    assert str(SignedIntegerType(15, PrimitiveType.CastMode.SATURATED)) == 'saturated int15'
    assert SignedIntegerType(64, PrimitiveType.CastMode.SATURATED).bit_length_set == {64}
    assert SignedIntegerType(8, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (-128, 127)  # type: ignore

    assert str(UnsignedIntegerType(15, PrimitiveType.CastMode.TRUNCATED)) == 'truncated uint15'
    assert UnsignedIntegerType(53, PrimitiveType.CastMode.SATURATED).bit_length_set == {53}
    assert UnsignedIntegerType(32, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (0,   # type: ignore
                                                                                               0xFFFFFFFF)

    assert str(FloatType(64, PrimitiveType.CastMode.SATURATED)) == 'saturated float64'
    assert FloatType(32, PrimitiveType.CastMode.SATURATED).bit_length_set == 32
    assert FloatType(16, PrimitiveType.CastMode.SATURATED).inclusive_value_range == (-65504, +65504)  # type: ignore

    rng = approx(-3.4028234664e+38), approx(+3.4028234664e+38)
    assert FloatType(32, PrimitiveType.CastMode.SATURATED).inclusive_value_range == rng  # type: ignore

    rng = approx(-1.7976931348623157e+308), approx(+1.7976931348623157e+308)
    assert FloatType(64, PrimitiveType.CastMode.SATURATED).inclusive_value_range == rng  # type: ignore

    with raises(InvalidBitLengthError):
        FloatType(8, PrimitiveType.CastMode.TRUNCATED)

    with raises(InvalidBitLengthError):
        SignedIntegerType(1, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        SignedIntegerType(0, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        UnsignedIntegerType(0, PrimitiveType.CastMode.SATURATED)

    with raises(InvalidBitLengthError):
        UnsignedIntegerType(65, PrimitiveType.CastMode.TRUNCATED)

    assert repr(SignedIntegerType(24, PrimitiveType.CastMode.SATURATED)) == \
        'SignedIntegerType(bit_length=24, cast_mode=<CastMode.SATURATED: 0>)'

    a = UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    b = BooleanType(PrimitiveType.CastMode.SATURATED)
    assert hash(a) != hash(b)
    assert hash(a) == hash(UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED))
    assert a == UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    assert b != UnsignedIntegerType(2, PrimitiveType.CastMode.TRUNCATED)
    assert a != b
    assert b == BooleanType(PrimitiveType.CastMode.SATURATED)
    assert b != 123    # Not implemented

    for bl in range(1, PrimitiveType.MAX_BIT_LENGTH + 1):
        if bl > 1:
            t = UnsignedIntegerType(bl, PrimitiveType.CastMode.SATURATED)  # type: PrimitiveType
        else:
            t = BooleanType(PrimitiveType.CastMode.SATURATED)
        assert t.standard_bit_length == (t.bit_length in {8, 16, 32, 64, 128, 256})



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import logging
from . import _serializable
from . import _expression
from . import _error
from . import _dsdl_definition
from . import _parser
from . import _data_schema_builder
from . import _port_id_ranges


class AssertionCheckFailureError(_error.InvalidDefinitionError):
    pass


class UndefinedDataTypeError(_error.InvalidDefinitionError):
    pass


class UndefinedIdentifierError(_error.InvalidDefinitionError):
    pass


class InvalidDirectiveError(_error.InvalidDefinitionError):
    pass


class UnregulatedFixedPortIDError(_error.InvalidDefinitionError):
    pass


class MissingSerializationModeError(_error.InvalidDefinitionError):
    pass


_logger = logging.getLogger(__name__)


class DataTypeBuilder(_parser.StatementStreamProcessor):
    def __init__(self,
                 definition:                      _dsdl_definition.DSDLDefinition,
                 lookup_definitions:              typing.Iterable[_dsdl_definition.DSDLDefinition],
                 print_output_handler:            typing.Callable[[int, str], None],
                 allow_unregulated_fixed_port_id: bool):
        self._definition = definition
        self._lookup_definitions = list(lookup_definitions)
        self._print_output_handler = print_output_handler
        self._allow_unregulated_fixed_port_id = allow_unregulated_fixed_port_id

        assert isinstance(self._definition, _dsdl_definition.DSDLDefinition)
        assert all(map(lambda x: isinstance(x, _dsdl_definition.DSDLDefinition), lookup_definitions))
        assert callable(self._print_output_handler)
        assert isinstance(self._allow_unregulated_fixed_port_id, bool)

        self._structs = [_data_schema_builder.DataSchemaBuilder()]
        self._is_deprecated = False

    def finalize(self) -> _serializable.CompositeType:
        if len(self._structs) == 1:     # Structure type
            builder, = self._structs     # type: _data_schema_builder.DataSchemaBuilder,
            out = self._make_composite(builder=builder,
                                       name=self._definition.full_name,
                                       version=self._definition.version,
                                       deprecated=self._is_deprecated,
                                       fixed_port_id=self._definition.fixed_port_id,
                                       source_file_path=self._definition.file_path,
                                       has_parent_service=False)
        else:  # Service type
            request_builder, response_builder = self._structs
            assert isinstance(request_builder, _data_schema_builder.DataSchemaBuilder)
            assert isinstance(response_builder, _data_schema_builder.DataSchemaBuilder)
            sep = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR
            request = self._make_composite(builder=request_builder,
                                           name=sep.join([self._definition.full_name, 'Request']),
                                           version=self._definition.version,
                                           deprecated=self._is_deprecated,
                                           fixed_port_id=None,
                                           source_file_path=self._definition.file_path,
                                           has_parent_service=True)
            response = self._make_composite(builder=response_builder,
                                            name=sep.join([self._definition.full_name, 'Response']),
                                            version=self._definition.version,
                                            deprecated=self._is_deprecated,
                                            fixed_port_id=None,
                                            source_file_path=self._definition.file_path,
                                            has_parent_service=True)
            # noinspection SpellCheckingInspection
            out = _serializable.ServiceType(                    # pozabito vse na svete
                request=request,                                # serdce zamerlo v grudi
                response=response,                              # tolko nebo tolko veter
                fixed_port_id=self._definition.fixed_port_id,   # tolko radost vperedi
            )

        assert isinstance(out, _serializable.CompositeType)
        if not self._allow_unregulated_fixed_port_id:
            port_id = out.fixed_port_id
            if port_id is not None:
                is_service_type = isinstance(out, _serializable.ServiceType)
                f = _port_id_ranges.is_valid_regulated_service_id if is_service_type else \
                    _port_id_ranges.is_valid_regulated_subject_id
                if not f(port_id, out.root_namespace):
                    raise UnregulatedFixedPortIDError(
                        'Regulated port ID %r for %s type %r is not valid. '
                        'Consider using allow_unregulated_fixed_port_id.' %
                        (port_id, 'service' if is_service_type else 'message', out.full_name))
        return out

    def on_constant(self,
                    constant_type: _serializable.SerializableType,
                    name: str,
                    value: _expression.Any) -> None:
        self._on_attribute()
        self._structs[-1].add_constant(_serializable.Constant(constant_type, name, value))

    def on_field(self, field_type: _serializable.SerializableType, name: str) -> None:
        self._on_attribute()
        self._structs[-1].add_field(_serializable.Field(field_type, name))

    def on_padding_field(self, padding_field_type: _serializable.VoidType) -> None:
        self._on_attribute()
        self._structs[-1].add_field(_serializable.PaddingField(padding_field_type))

    def on_directive(self,
                     line_number: int,
                     directive_name: str,
                     associated_expression_value: typing.Optional[_expression.Any]) -> None:
        try:
            handler = {
                'print':      self._on_print_directive,
                'assert':     self._on_assert_directive,
                'extent':     self._on_extent_directive,
                'sealed':     self._on_sealed_directive,
                'union':      self._on_union_directive,
                'deprecated': self._on_deprecated_directive,
            }[directive_name]
        except KeyError:
            raise InvalidDirectiveError('Unknown directive %r' % directive_name)
        else:
            assert callable(handler)
            return handler(line_number, associated_expression_value)

    def on_service_response_marker(self) -> None:
        if len(self._structs) > 1:
            raise _error.InvalidDefinitionError('Duplicated service response marker')

        self._structs.append(_data_schema_builder.DataSchemaBuilder())
        assert len(self._structs) == 2

    def resolve_top_level_identifier(self, name: str) -> _expression.Any:
        # Look only in the current data structure. The lookup cannot cross the service request/response boundary.
        for c in self._structs[-1].constants:
            if c.name == name:
                return c.value

        if name == '_offset_':
            bls = self._structs[-1].offset
            assert len(bls) > 0 and all(map(lambda x: isinstance(x, int), bls))
            return _expression.Set(map(_expression.Rational, bls))
        else:
            raise UndefinedIdentifierError('Undefined identifier: %r' % name)

    def resolve_versioned_data_type(self, name: str, version: _serializable.Version) -> _serializable.CompositeType:
        if _serializable.CompositeType.NAME_COMPONENT_SEPARATOR in name:
            full_name = name
        else:
            full_name = _serializable.CompositeType.NAME_COMPONENT_SEPARATOR.join([self._definition.full_namespace,
                                                                                   name])
            _logger.debug('The full name of a relatively referred type %r reconstructed as %r', name, full_name)

        del name
        found = list(filter(lambda d: d.full_name == full_name and d.version == version, self._lookup_definitions))
        if not found:
            # Play Sherlock to help the user with mistakes like https://forum.uavcan.org/t/dsdl-compilation-error/904/2
            requested_ns = full_name.split(_serializable.CompositeType.NAME_COMPONENT_SEPARATOR)[0]
            lookup_nss = set(x.root_namespace for x in self._lookup_definitions)
            subroot_ns = self._definition.name_components[1] if len(self._definition.name_components) > 2 else None
            error_description = 'Data type %s.%d.%d could not be found in the following root namespaces: %s. ' % \
                                (full_name, version.major, version.minor, lookup_nss or '(empty set)')
            if requested_ns not in lookup_nss and requested_ns == subroot_ns:
                error_description += ' Did you mean to use the directory %r instead of %r?' % (
                    os.path.join(self._definition.root_namespace_path, subroot_ns),
                    self._definition.root_namespace_path,
                )
            else:
                error_description += ' Please make sure that you specified the directories correctly.'
            raise UndefinedDataTypeError(error_description)

        if len(found) > 1:  # pragma: no cover
            raise _error.InternalError('Conflicting definitions: %r' % found)

        target_definition = found[0]
        assert isinstance(target_definition, _dsdl_definition.DSDLDefinition)
        assert target_definition.full_name == full_name
        assert target_definition.version == version
        # Recursion is cool.
        return target_definition.read(lookup_definitions=self._lookup_definitions,
                                      print_output_handler=self._print_output_handler,
                                      allow_unregulated_fixed_port_id=self._allow_unregulated_fixed_port_id)

    def _on_attribute(self) -> None:
        if isinstance(self._structs[-1].serialization_mode, _data_schema_builder.DelimitedSerializationMode):
            raise InvalidDirectiveError(
                'The extent directive can only be placed after the last attribute definition in the schema. '
                'This is to prevent errors if the extent is dependent on the bit length set of the data schema.'
            )

    def _on_print_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        _logger.info('Print directive at %s:%d%s', self._definition.file_path, line_number,
                     (': %s' % value) if value is not None else ' (no value to print)')
        self._print_output_handler(line_number, str(value if value is not None else ''))

    def _on_assert_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        if isinstance(value, _expression.Boolean):
            if not value.native_value:
                raise AssertionCheckFailureError('Assertion check has failed',
                                                 path=self._definition.file_path,
                                                 line=line_number)
            else:
                _logger.debug('Assertion check successful at %s:%d', self._definition.file_path, line_number)
        elif value is None:
            raise InvalidDirectiveError('Assert directive requires an expression')
        else:
            raise InvalidDirectiveError('The assertion check expression must yield a boolean, not %s' %
                                        value.TYPE_NAME)

    def _on_extent_directive(self, line_number: int, value: typing.Optional[_expression.Any]) -> None:
        if self._structs[-1].serialization_mode is not None:
            raise InvalidDirectiveError('Misplaced extent directive. The serialization mode is already set to %s' %
                                        self._structs[-1].serialization_mode)
        if value is None:
            raise InvalidDirectiveError('The extent directive requires an expression')
        elif isinstance(value, _expression.Rational):
            struct = self._structs[-1]
            bits = value.as_native_integer()
            struct.set_serialization_mode(_data_schema_builder.DelimitedSerializationMode(bits))
            _logger.debug('The extent is set to %d bits at %s:%d', bits, self._definition.file_path, line_number)
        else:
            raise InvalidDirectiveError('The extent directive expects a rational, not %s' % value.TYPE_NAME)

    def _on_sealed_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if self._structs[-1].serialization_mode is not None:
            raise InvalidDirectiveError('Misplaced sealing directive. The serialization mode is already set to %s' %
                                        self._structs[-1].serialization_mode)
        if value is not None:
            raise InvalidDirectiveError('The sealed directive does not expect an expression')
        self._structs[-1].set_serialization_mode(_data_schema_builder.SealedSerializationMode())

    def _on_union_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The union directive does not expect an expression')
        if self._structs[-1].union:
            raise InvalidDirectiveError('Duplicated union directive')
        if self._structs[-1].attributes:
            raise InvalidDirectiveError('The union directive must be placed before the first '
                                        'attribute definition')
        self._structs[-1].make_union()

    def _on_deprecated_directive(self, _ln: int, value: typing.Optional[_expression.Any]) -> None:
        if value is not None:
            raise InvalidDirectiveError('The deprecated directive does not expect an expression')
        if self._is_deprecated:
            raise InvalidDirectiveError('Duplicated deprecated directive')
        if len(self._structs) > 1:
            raise InvalidDirectiveError('The deprecated directive cannot be placed in the response section')
        if self._structs[-1].attributes:
            raise InvalidDirectiveError('The deprecated directive must be placed before the first '
                                        'attribute definition')
        self._is_deprecated = True

    @staticmethod
    def _make_composite(builder:            _data_schema_builder.DataSchemaBuilder,
                        name:               str,
                        version:            _serializable.Version,
                        deprecated:         bool,
                        fixed_port_id:      typing.Optional[int],
                        source_file_path:   str,
                        has_parent_service: bool) -> _serializable.CompositeType:
        ty = _serializable.UnionType if builder.union else _serializable.StructureType
        inner = ty(name=name,
                   version=version,
                   attributes=builder.attributes,
                   deprecated=deprecated,
                   fixed_port_id=fixed_port_id,
                   source_file_path=source_file_path,
                   has_parent_service=has_parent_service)  # type: _serializable.CompositeType
        sm = builder.serialization_mode
        if isinstance(sm, _data_schema_builder.DelimitedSerializationMode):
            out = _serializable.DelimitedType(inner, extent=sm.extent)  # type: _serializable.CompositeType
            _logger.debug('%r wrapped into %r', inner, out)
        elif isinstance(sm, _data_schema_builder.SealedSerializationMode):
            out = inner
        else:
            assert sm is None, 'I wish Python had a strong static type system'
            raise MissingSerializationModeError(
                '%s: Either `@sealed` or `@extent ...` are required. '
                'The smallest valid extent for this type (i.e., its max bit length) is %d bits (%d bytes). '
                'If you are not sure what this means, add the following line near the end of this definition: '
                '`@extent %d * 8`'
                % (inner.short_name, inner.extent, inner.extent // 8, DataTypeBuilder._suggest_extent_in_bytes(inner))
            )
        return out

    @staticmethod
    def _suggest_extent_in_bytes(model: _serializable.CompositeType) -> int:
        """
        An implementation-specific heuristic intended to lower the entry barrier.
        The numbers are subject to change between minor revisions.
        """
        return max(64, model.extent * 2 // 8)



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import fractions
from . import _any, _primitive, _container, _operator


# noinspection PyUnresolvedReferences,PyTypeChecker
def _unittest_expressions() -> None:
    r = _primitive.Rational
    s = _primitive.String

    for a in (True, False):
        for b in (True, False):
            assert _primitive.Boolean(a).native_value == a
            assert _operator.logical_not(_primitive.Boolean(a)).native_value == (not a)
            assert _operator.logical_and(_primitive.Boolean(a), _primitive.Boolean(b)).native_value == (a and b)
            assert _operator.logical_or(_primitive.Boolean(a), _primitive.Boolean(b)).native_value == (a or b)

    assert \
        _operator.equal(
            _operator.divide(
                _operator.multiply(
                    _operator.add(r(2), r(2)),
                    r(3)
                ),
                r(5)
            ),
            r(fractions.Fraction(12, 5))
        ).native_value

    assert _operator.add(s('123'), s('abc')).native_value == '123abc'  # type: ignore

    new_set = _operator.add(_container.Set([s('123'), s('456')]),
                            s('abc'))
    assert set(new_set) == {s('123abc'), s('456abc')}  # type: ignore

    new_set = _operator.add(s('abc'),
                            _container.Set([s('123'), s('456')]))
    assert set(new_set) == {s('abc123'), s('abc456')}  # type: ignore

    new_set = _operator.add(s('abc'),
                            _container.Set([_container.Set([s('123'), s('456')]),
                                            _container.Set([s('789'), s('987')])]))
    assert new_set == _container.Set([_container.Set([s('abc123'), s('abc456')]),
                                      _container.Set([s('abc789'), s('abc987')])])

    assert _operator.attribute(_container.Set([r(1), r(2), r(3), r(-4), r(-5)]), s('min')) == r(-5)
    assert _operator.attribute(_container.Set([r(1), r(2), r(3), r(-4), r(-5)]), s('max')) == r(3)


def _unittest_textual_representations() -> None:
    assert str(_primitive.Rational(fractions.Fraction(123, 456))) == '41/152'
    assert repr(_primitive.Rational(fractions.Fraction(123, 456))) == 'rational(41/152)'
    assert str(_primitive.Rational(-123)) == '-123'
    assert repr(_primitive.Rational(-123)) == 'rational(-123)'

    assert str(_primitive.Boolean(True)) == 'true'
    assert repr(_primitive.Boolean(False)) == 'bool(false)'

    assert str(_primitive.String('Hello\nworld!')) == r"'Hello\nworld!'"
    assert repr(_primitive.String('Hello\nworld!')) == r"string('Hello\nworld!')"

    tmp = str(_container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]))
    assert tmp == '{1, -9/7}' or tmp == '{-9/7, 1}'

    tmp = repr(_container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]))
    assert tmp == 'set({1, -9/7})' or tmp == 'set({-9/7, 1})'

    tmp = str(_container.Set([_container.Set([_primitive.Rational(1), _primitive.Rational(fractions.Fraction(-9, 7))]),
                              _container.Set([_primitive.Rational(fractions.Fraction(90, 7))])]))
    assert \
        tmp == '{{1, -9/7}, {90/7}}' or \
        tmp == '{{-9/7, 1}, {90/7}}' or \
        tmp == '{{90/7}, {-9/7, 1}}' or \
        tmp == '{{90/7}, {1, -9/7}}'

    assert repr(_container.Set([_primitive.String('123')])) == "set({'123'})"


# noinspection PyTypeChecker
def _unittest_basic() -> None:
    from pytest import raises

    assert hash(_primitive.Boolean(True)) == hash(True)
    assert _primitive.Boolean(True) == _primitive.Boolean(True)
    assert _primitive.Boolean(True) != _primitive.Boolean(False)
    assert _primitive.Boolean(True) != _primitive.Rational(1)         # sic!
    assert _primitive.Boolean(True) != _primitive.Rational(123)
    assert _primitive.Boolean(True) != _container.Set([_primitive.Boolean(True)])

    with raises(ValueError):
        _primitive.Boolean(int)       # type: ignore

    with raises(ValueError):
        _primitive.Rational({123})    # type: ignore

    with raises(ValueError):
        _primitive.Rational('123')    # type: ignore

    with raises(ValueError):
        _primitive.String(123)       # type: ignore

    with raises(ValueError):
        _container.Set([123])        # type: ignore

    assert _primitive.Rational(123).is_integer()
    assert not _primitive.Rational(fractions.Fraction(123, 124)).is_integer()
    assert _primitive.Rational(-123).as_native_integer() == -123
    with raises(_any.InvalidOperandError):
        _primitive.Rational(fractions.Fraction(123, 124)).as_native_integer()



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

from ._serializable import SerializableType as SerializableType

from ._primitive import PrimitiveType as PrimitiveType
from ._primitive import BooleanType as BooleanType
from ._primitive import FloatType as FloatType
from ._primitive import ValueRange as ValueRange
from ._primitive import ArithmeticType as ArithmeticType
from ._primitive import IntegerType as IntegerType
from ._primitive import SignedIntegerType as SignedIntegerType
from ._primitive import UnsignedIntegerType as UnsignedIntegerType

from ._void import VoidType as VoidType

from ._array import ArrayType as ArrayType
from ._array import FixedLengthArrayType as FixedLengthArrayType
from ._array import VariableLengthArrayType as VariableLengthArrayType

from ._composite import CompositeType as CompositeType
from ._composite import UnionType as UnionType
from ._composite import StructureType as StructureType
from ._composite import DelimitedType as DelimitedType
from ._composite import ServiceType as ServiceType
from ._composite import Version as Version

from ._attribute import Attribute as Attribute
from ._attribute import Field as Field
from ._attribute import PaddingField as PaddingField
from ._attribute import Constant as Constant



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing
from . import _error
from . import _serializable
from . import _bit_length_set


class BitLengthAnalysisError(_error.InvalidDefinitionError):
    pass


class SerializationMode:
    """Serialization mode: either delimited (with extent) or sealed."""
    def __str__(self) -> str:
        raise NotImplementedError


class DelimitedSerializationMode(SerializationMode):
    def __init__(self, extent: int):
        self.extent = int(extent)

    def __str__(self) -> str:
        return 'delimited (extent %d bits)' % self.extent


class SealedSerializationMode(SerializationMode):
    def __str__(self) -> str:
        return 'sealed'


class DataSchemaBuilder:
    def __init__(self) -> None:
        self._fields = []                   # type: typing.List[_serializable.Field]
        self._constants = []                # type: typing.List[_serializable.Constant]
        self._serialization_mode = None     # type: typing.Optional[SerializationMode]
        self._is_union = False
        self._bit_length_computed_at_least_once = False

    @property
    def fields(self) -> typing.List[_serializable.Field]:
        assert all(map(lambda x: isinstance(x, _serializable.Field), self._fields))
        return self._fields

    @property
    def constants(self) -> typing.List[_serializable.Constant]:
        assert all(map(lambda x: isinstance(x, _serializable.Constant), self._constants))
        return self._constants

    @property
    def attributes(self) -> typing.List[_serializable.Attribute]:  # noinspection PyTypeChecker
        out = []  # type: typing.List[_serializable.Attribute]
        out += self.fields
        out += self.constants
        return out

    @property
    def serialization_mode(self) -> typing.Optional[SerializationMode]:
        return self._serialization_mode

    @property
    def union(self) -> bool:
        return self._is_union

    @property
    def offset(self) -> _bit_length_set.BitLengthSet:
        # We set this flag in order to detect invalid reliance on the bit length estimates for unions:
        # we process definitions sequentially, statement-by-statement, so we can't know if there are going to be
        # extra fields added after the bit length values are computed. If we are building a regular structure,
        # this is fine, because in that case each computed value refers to the offset of the next field (if there
        # is one) or the total length of the structure (if we're past the last field). With unions, however,
        # there is no concept of inter-field offset because a union holds exactly one field at any moment;
        # only the total offset (i.e., total size) is defined.
        self._bit_length_computed_at_least_once = True
        ty = _serializable.UnionType if self.union else _serializable.StructureType
        out = ty.aggregate_bit_length_sets([f.data_type for f in self.fields])  # type: ignore
        assert isinstance(out, _bit_length_set.BitLengthSet) and len(out) > 0
        return out

    def add_field(self, field: _serializable.Field) -> None:
        if self.union and self._bit_length_computed_at_least_once:
            # Refer to the DSDL specification for the background information.
            raise BitLengthAnalysisError('Inter-field offset is not defined for unions; '
                                         'previously performed bit length analysis is invalid')
        assert isinstance(field, _serializable.Field)
        self._fields.append(field)

    def add_constant(self, constant: _serializable.Constant) -> None:
        assert isinstance(constant, _serializable.Constant)
        self._constants.append(constant)

    def set_serialization_mode(self, mode: SerializationMode) -> None:
        assert self._serialization_mode is None
        self._serialization_mode = mode

    def make_union(self) -> None:
        assert not self.union
        self._is_union = True



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
from .. import _error


class InvalidOperandError(_error.InvalidDefinitionError):
    pass


class UndefinedOperatorError(InvalidOperandError):
    """Thrown when there is no matching operator for the supplied arguments."""
    def __init__(self) -> None:
        super(UndefinedOperatorError, self).__init__('The requested operator is not defined for the provided arguments')


class UndefinedAttributeError(InvalidOperandError):
    """Thrown when the requested attribute does not exist."""
    def __init__(self) -> None:
        super(UndefinedAttributeError, self).__init__('Invalid attribute name')


class Any(abc.ABC):
    """
    This abstract class represents an arbitrary intrinsic DSDL expression value.
    Both serializable types and expression types derive from this common ancestor.

    Per the DSDL data model, a serializable type is also a value.
    Serializable types have the suffix ``Type`` because their instances represent not DSDL values but DSDL types.

    Instances of this type can be pickled.
    """

    TYPE_NAME = None    # type: str
    """
    The DSDL-name of the data type implemented by the class, as defined in Specification.
    """

    @abc.abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def __eq__(self, other: object) -> bool:
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def __str__(self) -> str:
        """Returns a DSDL spec-compatible textual representation of the contained value suitable for printing."""
        raise NotImplementedError  # pragma: no cover

    def __repr__(self) -> str:
        return self.TYPE_NAME + '(' + str(self) + ')'

    # Unary operators.
    def _logical_not(self) -> 'Boolean': raise UndefinedOperatorError

    def _positive(self) -> 'Any': raise UndefinedOperatorError

    def _negative(self) -> 'Any': raise UndefinedOperatorError

    # Binary operators.
    # The types of the operators defined here must match the specification.
    # Make sure to use least generic types in the derived classes - Python allows covariant return types.
    def _logical_or(self, right: 'Any')  -> 'Boolean': raise UndefinedOperatorError
    def _logical_and(self, right: 'Any') -> 'Boolean': raise UndefinedOperatorError

    def _equal(self, right: 'Any')            -> 'Boolean': raise UndefinedOperatorError  # pragma: no branch
    def _less_or_equal(self, right: 'Any')    -> 'Boolean': raise UndefinedOperatorError
    def _greater_or_equal(self, right: 'Any') -> 'Boolean': raise UndefinedOperatorError
    def _less(self, right: 'Any')             -> 'Boolean': raise UndefinedOperatorError
    def _greater(self, right: 'Any')          -> 'Boolean': raise UndefinedOperatorError

    def _bitwise_or(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_or_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _bitwise_xor(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_xor_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _bitwise_and(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _bitwise_and_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _add(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _add_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _subtract(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _subtract_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _multiply(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _multiply_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _divide(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _divide_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _modulo(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _modulo_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    def _power(self, right: 'Any')      -> 'Any': raise UndefinedOperatorError
    def _power_right(self, left: 'Any') -> 'Any': raise UndefinedOperatorError

    # Attribute access operator. It is a binary operator as well, but its semantics is slightly different.
    # Implementations must invoke super()._attribute() when they encounter an unknown attribute, to allow
    # the parent classes to handle the requested attribute as a fallback option.
    def _attribute(self, name: 'String') -> 'Any': raise UndefinedAttributeError


# This import must be located at the bottom to break the circular dependency in the type annotations above.
# We must import specific names as opposed to the whole module because the latter breaks MyPy.
from ._primitive import Boolean, String    # noqa



#
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import abc
import math
import typing
from .._bit_length_set import BitLengthSet
from ._serializable import SerializableType, TypeParameterError
from ._primitive import UnsignedIntegerType, PrimitiveType


class InvalidNumberOfElementsError(TypeParameterError):
    pass


class ArrayType(SerializableType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(ArrayType, self).__init__()
        self._element_type = element_type
        self._capacity = int(capacity)
        if self._capacity < 1:
            raise InvalidNumberOfElementsError('Array capacity cannot be less than 1')

    @property
    def element_type(self) -> SerializableType:
        return self._element_type

    @property
    def capacity(self) -> int:
        """
        The (maximum) number of elements in the (variable-length) array.
        """
        return self._capacity

    @property
    def string_like(self) -> bool:
        """
        True if the array might contain a text string, in which case it is termed to be "string-like".
        A string-like array is a variable-length array of ``uint8``.
        See https://github.com/UAVCAN/specification/issues/51.
        """
        return False

    @property
    def alignment_requirement(self) -> int:
        """
        The alignment requirement of an array equals that of its element type.
        The length of the serialized representation of any type is a multiple of its alignment requirement;
        therefore, every element is always placed such that its alignment requirement is satisfied.
        """
        return self.element_type.alignment_requirement

    @abc.abstractmethod
    def __str__(self) -> str:   # pragma: no cover
        raise NotImplementedError


class FixedLengthArrayType(ArrayType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(FixedLengthArrayType, self).__init__(element_type, capacity)

    @property
    def bit_length_set(self) -> BitLengthSet:
        # This can be further generalized as a Cartesian product of the element type's bit length set taken N times,
        # where N is the capacity of the array. However, we avoid such generalization because it leads to a mild
        # combinatorial explosion even with small arrays, resorting to this special case instead. The difference in
        # performance measured on the standard data type set was about tenfold.
        return self.element_type.bit_length_set.elementwise_sum_k_multicombinations(self.capacity).\
            pad_to_alignment(self.alignment_requirement)

    def enumerate_elements_with_offsets(self, base_offset: typing.Optional[BitLengthSet] = None) \
            -> typing.Iterator[typing.Tuple[int, BitLengthSet]]:
        """
        This is a convenience method for code generation.
        Its behavior mimics that of :meth:`pydsdl.StructureType.iterate_fields_with_offsets`,
        except that we iterate over indexes instead of fields.

        :param base_offset: The base offset to add to each element. If not supplied, assumed to be ``{0}``.
            The base offset will be implicitly padded out to :attr:`alignment_requirement`.

        :returns: For an N-element array, an iterator over N elements, where each element is a tuple of the index
            of the array element (zero-based) and its offset as a bit length set.
        """
        base_offset = BitLengthSet(base_offset or 0).pad_to_alignment(self.alignment_requirement)
        _self_test_base_offset = BitLengthSet(0)
        for index in range(self.capacity):
            assert base_offset.is_aligned_at(self.element_type.alignment_requirement),\
                'The bit length set of the element type computed incorrectly: length % alignment = 0 does not hold.'
            yield index, BitLengthSet(base_offset)      # We yield a copy of the offset to prevent mutation
            base_offset += self.element_type.bit_length_set

            # This is only for ensuring that the logic is functioning as intended.
            # Combinatorial transformations are easy to mess up, so we have to employ defensive programming.
            assert self.element_type.bit_length_set.elementwise_sum_k_multicombinations(index) == _self_test_base_offset
            _self_test_base_offset += self.element_type.bit_length_set

    def __str__(self) -> str:
        return '%s[%d]' % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return 'FixedLengthArrayType(element_type=%r, capacity=%r)' % (self.element_type, self.capacity)


def _unittest_fixed_array() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    su8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    ti64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(FixedLengthArrayType(su8, 4)) == 'truncated uint8[4]'
    assert str(FixedLengthArrayType(ti64, 1)) == 'saturated int64[1]'

    assert not FixedLengthArrayType(su8, 4).string_like
    assert not FixedLengthArrayType(ti64, 1).string_like

    assert FixedLengthArrayType(su8, 4).bit_length_set == 32
    assert FixedLengthArrayType(su8, 200).capacity == 200
    assert FixedLengthArrayType(ti64, 200).element_type is ti64

    with raises(InvalidNumberOfElementsError):
        FixedLengthArrayType(ti64, 0)

    assert repr(FixedLengthArrayType(ti64, 128)) == \
        'FixedLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), ' \
        'capacity=128)'

    small = FixedLengthArrayType(su8, 2)
    assert small.bit_length_set == {16}
    assert list(small.enumerate_elements_with_offsets()) == [(0, BitLengthSet(0)), (1, BitLengthSet(8))]


class VariableLengthArrayType(ArrayType):
    def __init__(self,
                 element_type: SerializableType,
                 capacity: int):
        super(VariableLengthArrayType, self).__init__(element_type, capacity)

        # Construct the implicit array length prefix type.
        length_field_length = 2 ** math.ceil(math.log2(max(self.BITS_PER_BYTE, self.capacity.bit_length())))

        # If the length field is less than the alignment requirement (which, at the time of writing this,
        # is not possible because the max alignment is 8 and the min length length is also 8),
        # it would break the alignment of the array elements. Hence, we ensure that it is never smaller.
        length_field_length = max(length_field_length, self.alignment_requirement)
        assert length_field_length % self.element_type.alignment_requirement == 0

        self._length_field_type = UnsignedIntegerType(length_field_length, PrimitiveType.CastMode.TRUNCATED)

    @property
    def bit_length_set(self) -> BitLengthSet:
        # Can't use @cached_property because it is unavailable before Python 3.8 and it breaks Sphinx and MyPy.
        # Caching is important because bit length set derivation is a very expensive operation.
        att = '_8467150963'
        if not hasattr(self, att):
            setattr(self, att, self._compute_bit_length_set())
        out = getattr(self, att)
        assert isinstance(out, BitLengthSet)
        return out

    @property
    def string_like(self) -> bool:
        """See the base class."""
        et = self.element_type      # Without this temporary MyPy yields a false positive type error
        return isinstance(et, UnsignedIntegerType) and (et.bit_length == self.BITS_PER_BYTE)

    @property
    def length_field_type(self) -> UnsignedIntegerType:
        """
        The unsigned integer type of the implicit array length field.
        Note that the set of valid length values is a subset of that of the returned type.
        """
        assert self._length_field_type.bit_length % self.element_type.alignment_requirement == 0
        return self._length_field_type

    def _compute_bit_length_set(self) -> BitLengthSet:
        # Please refer to the corresponding implementation for the fixed-length array.
        # The idea here is that we treat the variable-length array as a combination of fixed-length arrays of
        # different sizes, from zero elements up to the maximum number of elements.
        output = BitLengthSet()
        for capacity in range(self.capacity + 1):
            output |= self.element_type.bit_length_set.elementwise_sum_k_multicombinations(capacity)
        output += self.length_field_type.bit_length
        return output.pad_to_alignment(self.alignment_requirement)

    def __str__(self) -> str:
        return '%s[<=%d]' % (self.element_type, self.capacity)

    def __repr__(self) -> str:
        return 'VariableLengthArrayType(element_type=%r, capacity=%r)' % (self.element_type, self.capacity)


def _unittest_variable_array() -> None:
    from pytest import raises
    from ._primitive import SignedIntegerType

    tu8 = UnsignedIntegerType(8, cast_mode=PrimitiveType.CastMode.TRUNCATED)
    si64 = SignedIntegerType(64, cast_mode=PrimitiveType.CastMode.SATURATED)

    assert str(VariableLengthArrayType(tu8, 4)) == 'truncated uint8[<=4]'
    assert str(VariableLengthArrayType(si64, 255)) == 'saturated int64[<=255]'

    assert VariableLengthArrayType(tu8, 4).string_like
    assert not VariableLengthArrayType(si64, 1).string_like

    # Mind the length prefix!
    assert VariableLengthArrayType(tu8, 3).bit_length_set == {8, 16, 24, 32}
    assert VariableLengthArrayType(tu8, 1).bit_length_set == {8, 16}
    assert max(VariableLengthArrayType(tu8, 255).bit_length_set) == 2048

    assert VariableLengthArrayType(tu8, 200).capacity == 200
    assert VariableLengthArrayType(tu8, 200).element_type is tu8

    with raises(InvalidNumberOfElementsError):
        VariableLengthArrayType(si64, 0)

    assert repr(VariableLengthArrayType(si64, 128)) == \
        'VariableLengthArrayType(element_type=SignedIntegerType(bit_length=64, cast_mode=<CastMode.SATURATED: 0>), ' \
        'capacity=128)'

    small = VariableLengthArrayType(tu8, 2)
    assert small.bit_length_set == {8, 16, 24}

    outer = FixedLengthArrayType(small, 2)
    assert outer.bit_length_set == {16, 24, 32, 40, 48}

    assert VariableLengthArrayType(tu8, 100).length_field_type.bit_length == 8
    assert VariableLengthArrayType(tu8, 10000).length_field_type.bit_length == 16
    assert VariableLengthArrayType(tu8, 1000000).length_field_type.bit_length == 32
    assert VariableLengthArrayType(tu8, 10000000000).length_field_type.bit_length == 64



#
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import os
import typing
import logging
import fnmatch
import collections
from . import _serializable
from . import _dsdl_definition
from . import _error


class RootNamespaceNameCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there is more than one namespace under the same name.
    This may occur if there are identically named namespaces located in different directories.
    """
    pass


class DataTypeNameCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there are conflicting data type names.
    """
    pass


class NestedRootNamespaceError(_error.InvalidDefinitionError):
    """
    Nested root namespaces are not allowed. This exception is thrown when this rule is violated.
    """
    pass


class FixedPortIDCollisionError(_error.InvalidDefinitionError):
    """
    Raised when there is more than one data type, or different major versions of the same data type
    using the same fixed port ID.
    """
    pass


class MultipleDefinitionsUnderSameVersionError(_error.InvalidDefinitionError):
    """
    For example::

        Type.1.0.uavcan
        2800.Type.1.0.uavcan
        2801.Type.1.0.uavcan
    """
    pass


class VersionsOfDifferentKindError(_error.InvalidDefinitionError):
    """
    Definitions that share the same name but are of different kinds.
    """
    pass


class MinorVersionFixedPortIDError(_error.InvalidDefinitionError):
    """
    Different fixed port-ID under the same major version, or a fixed port ID was removed under the same major version.
    """
    pass


class ExtentConsistencyError(_error.InvalidDefinitionError):
    """
    Different extent under the same major version.
    """
    pass


class SealingConsistencyError(_error.InvalidDefinitionError):
    """
    Different sealing status under the same major version.
    """
    pass


PrintOutputHandler = typing.Callable[[str, int, str], None]
"""Invoked when the frontend encounters a print directive or needs to output a generic diagnostic."""


def read_namespace(root_namespace_directory:        str,
                   lookup_directories:              typing.Optional[typing.Union[str, typing.Iterable[str]]] = None,
                   print_output_handler:            typing.Optional[PrintOutputHandler] = None,
                   allow_unregulated_fixed_port_id: bool = False) -> \
        typing.List[_serializable.CompositeType]:
    """
    This function is the main entry point of the library.
    It reads all DSDL definitions from the specified root namespace directory and produces the annotated AST.

    :param root_namespace_directory: The path of the root namespace directory that will be read.
        For example, ``dsdl/uavcan`` to read the ``uavcan`` namespace.

    :param lookup_directories: List of other namespace directories containing data type definitions that are
        referred to from the target root namespace. For example, if you are reading a vendor-specific namespace,
        the list of lookup directories should always include a path to the standard root namespace ``uavcan``,
        otherwise the types defined in the vendor-specific namespace won't be able to use data types from the
        standard namespace.

    :param print_output_handler: If provided, this callable will be invoked when a ``@print`` directive
        is encountered or when the frontend needs to emit a diagnostic;
        the arguments are: path, line number (1-based), text.
        If not provided, no output will be produced except for the standard Python logging subsystem
        (but ``@print`` expressions will be evaluated anyway, and a failed evaluation will be a fatal error).

    :param allow_unregulated_fixed_port_id: Do not reject unregulated fixed port identifiers.
        As demanded by the specification, the frontend rejects unregulated fixed port ID by default.
        This is a dangerous feature that must not be used unless you understand the risks.
        Please read https://uavcan.org/guide.

    :return: A list of :class:`pydsdl.CompositeType` sorted lexicographically by full data type name,
             then by major version (newest version first), then by minor version (newest version first).
             The ordering guarantee allows the caller to always find the newest version simply by picking
             the first matching occurrence.

    :raises: :class:`pydsdl.FrontendError`, :class:`MemoryError`, :class:`SystemError`,
        :class:`OSError` if directories do not exist or inaccessible,
        :class:`ValueError`/:class:`TypeError` if the arguments are invalid.
    """
    # Add the own root namespace to the set of lookup directories, sort lexicographically, remove duplicates.
    # We'd like this to be an iterable list of strings but we handle the common practice of passing in a single path.
    if lookup_directories is None:
        lookup_directories_path_list = []  # type: typing.Iterable[str]
    elif isinstance(lookup_directories, (str, bytes)):
        lookup_directories_path_list = [lookup_directories]
    else:
        lookup_directories_path_list = lookup_directories

    for a in lookup_directories_path_list:
        if not isinstance(a, str):  # non-string paths
            raise TypeError('Lookup directories shall be an iterable of strings. Found in list: ' + type(a).__name__)
        _logger.debug(_LOG_LIST_ITEM_PREFIX + a)

    lookup_directories_sorted_path_list = list(
        sorted(set(list(lookup_directories_path_list) + [root_namespace_directory])))

    # Normalize paths.
    root_namespace_directory = os.path.abspath(root_namespace_directory)
    lookup_directories_sorted_path_list = list(
        map(lambda d: str(os.path.abspath(d)), lookup_directories_sorted_path_list))
    _logger.debug('Lookup directories are listed below:')
    for a in lookup_directories_sorted_path_list:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + a)

    # Check the namespaces.
    _ensure_no_nested_root_namespaces(lookup_directories_sorted_path_list)
    _ensure_no_namespace_name_collisions(lookup_directories_sorted_path_list)

    # Construct DSDL definitions from the target and the lookup dirs.
    target_dsdl_definitions = _construct_dsdl_definitions_from_namespace(root_namespace_directory)
    if not target_dsdl_definitions:
        _logger.info('The namespace at %r is empty', root_namespace_directory)
        return []
    _logger.debug('Target DSDL definitions are listed below:')
    for x in target_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    lookup_dsdl_definitions = []    # type: typing.List[_dsdl_definition.DSDLDefinition]
    for ld in lookup_directories_sorted_path_list:
        lookup_dsdl_definitions += _construct_dsdl_definitions_from_namespace(ld)

    # Check for collisions against the lookup definitions also.
    _ensure_no_name_collisions(target_dsdl_definitions,
                               lookup_dsdl_definitions)

    _logger.debug('Lookup DSDL definitions are listed below:')
    for x in lookup_dsdl_definitions:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + str(x))

    _logger.info('Reading %d definitions from the root namespace %r, '
                 'with %d lookup definitions located in root namespaces: %s',
                 len(target_dsdl_definitions),
                 list(set(map(lambda t: t.root_namespace, target_dsdl_definitions)))[0],
                 len(lookup_dsdl_definitions),
                 ', '.join(set(sorted(map(lambda t: t.root_namespace, lookup_dsdl_definitions)))))

    # Read the constructed definitions.
    types = _read_namespace_definitions(target_dsdl_definitions,
                                        lookup_dsdl_definitions,
                                        print_output_handler,
                                        allow_unregulated_fixed_port_id)

    # Note that we check for collisions in the read namespace only.
    # We intentionally ignore (do not check for) possible collisions in the lookup directories,
    # because that would exceed the expected scope of responsibility of the frontend, and the lookup
    # directories may contain issues and mistakes that are outside of the control of the user (e.g.,
    # they could be managed by a third party) -- the user shouldn't be affected by mistakes committed
    # by the third party.
    _ensure_no_fixed_port_id_collisions(types)
    _ensure_minor_version_compatibility(types)

    return types


_DSDL_FILE_GLOB = '*.uavcan'
_LOG_LIST_ITEM_PREFIX = ' ' * 4

_logger = logging.getLogger(__name__)


def _read_namespace_definitions(target_definitions:              typing.List[_dsdl_definition.DSDLDefinition],
                                lookup_definitions:              typing.List[_dsdl_definition.DSDLDefinition],
                                print_output_handler:            typing.Optional[PrintOutputHandler] = None,
                                allow_unregulated_fixed_port_id: bool = False) -> \
        typing.List[_serializable.CompositeType]:
    """
    Construct type descriptors from the specified target definitions.
    Allow the target definitions to use the lookup definitions within themselves.
    :param target_definitions:  Which definitions to read.
    :param lookup_definitions:  Which definitions can be used by the processed definitions.
    :return: A list of types.
    """
    def make_print_handler(definition: _dsdl_definition.DSDLDefinition) -> typing.Callable[[int, str], None]:
        def handler(line_number: int, text: str) -> None:
            if print_output_handler:  # pragma: no branch
                assert isinstance(line_number, int) and isinstance(text, str)
                assert line_number > 0, 'Line numbers must be one-based'
                print_output_handler(definition.file_path, line_number, text)
        return handler

    types = []  # type: typing.List[_serializable.CompositeType]
    for tdd in target_definitions:
        try:
            dt = tdd.read(lookup_definitions,
                          make_print_handler(tdd),
                          allow_unregulated_fixed_port_id)
        except _error.FrontendError as ex:    # pragma: no cover
            ex.set_error_location_if_unknown(path=tdd.file_path)
            raise ex
        except (MemoryError, SystemError):    # pragma: no cover
            raise
        except Exception as ex:     # pragma: no cover
            raise _error.InternalError(culprit=ex, path=tdd.file_path) from ex
        else:
            types.append(dt)

    return types


def _ensure_no_name_collisions(target_definitions: typing.List[_dsdl_definition.DSDLDefinition],
                               lookup_definitions: typing.List[_dsdl_definition.DSDLDefinition]) -> None:
    for tg in target_definitions:
        for lu in lookup_definitions:
            if tg.full_name != lu.full_name and tg.full_name.lower() == lu.full_name.lower():
                raise DataTypeNameCollisionError('Full name of this definition differs from %r only by letter case, '
                                                 'which is not permitted' % lu.file_path, path=tg.file_path)

            if tg.full_namespace.lower().startswith(lu.full_name.lower()):  # pragma: no cover
                raise DataTypeNameCollisionError('The namespace of this type conflicts with %r' % lu.file_path,
                                                 path=tg.file_path)

            if lu.full_namespace.lower().startswith(tg.full_name.lower()):
                raise DataTypeNameCollisionError('This type conflicts with the namespace of %r' % lu.file_path,
                                                 path=tg.file_path)


def _ensure_no_fixed_port_id_collisions(types: typing.List[_serializable.CompositeType]) -> None:
    for a in types:
        for b in types:
            different_names = a.full_name != b.full_name
            different_major_versions = a.version.major != b.version.major
            # Must be the same kind because port ID sets of subjects and services are orthogonal
            same_kind = isinstance(a, _serializable.ServiceType) == isinstance(b, _serializable.ServiceType)
            # Data types where the major version is zero are allowed to collide
            both_released = (a.version.major > 0) and (b.version.major > 0)

            fpid_must_be_different = same_kind and (different_names or (different_major_versions and both_released))

            if fpid_must_be_different:
                if a.has_fixed_port_id and b.has_fixed_port_id:
                    if a.fixed_port_id == b.fixed_port_id:
                        raise FixedPortIDCollisionError(
                            'The fixed port ID of this definition is also used in %r' % b.source_file_path,
                            path=a.source_file_path
                        )


def _ensure_minor_version_compatibility(types: typing.List[_serializable.CompositeType]) -> None:
    by_name = collections.defaultdict(list)  # type: typing.DefaultDict[str, typing.List[_serializable.CompositeType]]
    for t in types:
        by_name[t.full_name].append(t)

    for definitions in by_name.values():
        by_major = \
            collections.defaultdict(list)  # type: typing.DefaultDict[int, typing.List[_serializable.CompositeType]]
        for t in definitions:
            by_major[t.version.major].append(t)

        for subject_to_check in by_major.values():
            _logger.debug('Minor version compatibility check amongst: %s', [str(x) for x in subject_to_check])
            for a in subject_to_check:
                for b in subject_to_check:
                    if a is not b:
                        _ensure_minor_version_compatibility_pairwise(a, b)


def _ensure_minor_version_compatibility_pairwise(a: _serializable.CompositeType,
                                                 b: _serializable.CompositeType) -> None:
    assert a is not b
    assert a.version.major == b.version.major
    assert a.full_name == b.full_name

    # Version collision
    if a.version.minor == b.version.minor:
        raise MultipleDefinitionsUnderSameVersionError(
            'This definition shares its version number with %r' % b.source_file_path,
            path=a.source_file_path
        )

    # Must be of the same kind: both messages or both services
    if isinstance(a, _serializable.ServiceType) != isinstance(b, _serializable.ServiceType):
        raise VersionsOfDifferentKindError(
            'This definition is not of the same kind as %r' % b.source_file_path,
            path=a.source_file_path
        )

    # Must use either the same RPID, or the older one should not have an RPID
    if a.has_fixed_port_id == b.has_fixed_port_id:
        if a.fixed_port_id != b.fixed_port_id:
            raise MinorVersionFixedPortIDError(
                'Different fixed port ID values under the same version %r' % b.source_file_path,
                path=a.source_file_path
            )
    else:
        must_have = a if a.version.minor > b.version.minor else b
        if not must_have.has_fixed_port_id:
            raise MinorVersionFixedPortIDError(
                'Fixed port ID cannot be removed under the same major version',
                path=must_have.source_file_path
            )

    # Extent and sealing equality
    if isinstance(a, _serializable.ServiceType) and isinstance(b, _serializable.ServiceType):
        _ensure_minor_version_compatibility_pairwise(a.request_type, b.request_type)
        _ensure_minor_version_compatibility_pairwise(a.response_type, b.response_type)
    else:
        if a.extent != b.extent:
            raise ExtentConsistencyError(
                'The extent of %s is %d bits, whereas the extent of %s is %d bits. '
                'The types share the same major version, so their extents should be equal '
                'to avoid wire compatibility issues.' %
                (
                    a, a.extent,
                    b, b.extent
                ),
                path=a.source_file_path
            )
        a_sealed = not isinstance(a, _serializable.DelimitedType)
        b_sealed = not isinstance(b, _serializable.DelimitedType)
        sealing_name = ['delimited', 'sealed']
        if a_sealed != b_sealed:
            raise SealingConsistencyError(
                '%s is %s, but %s is %s. '
                'Mixing sealed and delimited types under the same major version will cause wire compatibility issues.' %
                (
                    a, sealing_name[a_sealed],
                    b, sealing_name[b_sealed],
                ),
                path=a.source_file_path
            )


def _ensure_no_nested_root_namespaces(directories: typing.Iterable[str]) -> None:
    directories = list(sorted([str(os.path.abspath(x)) for x in set(directories)]))
    for a in directories:
        for b in directories:
            if (a != b) and a.startswith(b):
                raise NestedRootNamespaceError(
                    'The following namespace is nested inside this one, which is not permitted: %r' % a,
                    path=b
                )


def _ensure_no_namespace_name_collisions(directories: typing.Iterable[str]) -> None:
    def get_namespace_name(d: str) -> str:
        return os.path.split(d)[-1]

    directories = list(sorted([str(os.path.abspath(x)) for x in set(directories)]))
    for a in directories:
        for b in directories:
            if (a != b) and get_namespace_name(a).lower() == get_namespace_name(b).lower():
                raise RootNamespaceNameCollisionError('The name of this namespace conflicts with %r' % b, path=a)


def _construct_dsdl_definitions_from_namespace(root_namespace_path: str) \
        -> typing.List[_dsdl_definition.DSDLDefinition]:
    """
    Accepts a directory path, returns a sorted list of abstract DSDL file representations. Those can be read later.
    The definitions are sorted by name lexicographically, then by major version (greatest version first),
    then by minor version (same ordering as the major version).
    """
    def on_walk_error(os_ex: Exception) -> None:
        raise os_ex     # pragma: no cover

    walker = os.walk(root_namespace_path,
                     onerror=on_walk_error,
                     followlinks=True)

    source_file_paths = []  # type: typing.List[str]
    for root, _dirnames, filenames in walker:
        for filename in fnmatch.filter(filenames, _DSDL_FILE_GLOB):
            source_file_paths.append(os.path.join(root, filename))

    _logger.debug('DSDL files in the namespace dir %r are listed below:', root_namespace_path)
    for a in source_file_paths:
        _logger.debug(_LOG_LIST_ITEM_PREFIX + a)

    output = []  # type: typing.List[_dsdl_definition.DSDLDefinition]
    for fp in source_file_paths:
        dsdl_def = _dsdl_definition.DSDLDefinition(fp, root_namespace_path)
        output.append(dsdl_def)

    # Lexicographically by name, newest version first.
    return list(sorted(output, key=lambda d: (d.full_name, -d.version.major, -d.version.minor)))


def _unittest_dsdl_definition_constructor() -> None:
    import tempfile
    from ._dsdl_definition import FileNameFormatError

    directory = tempfile.TemporaryDirectory()
    root_ns_dir = os.path.join(directory.name, 'foo')

    os.mkdir(root_ns_dir)
    os.mkdir(os.path.join(root_ns_dir, 'nested'))

    def touchy(relative_path: str) -> None:
        p = os.path.join(root_ns_dir, relative_path.replace('/', os.path.sep))
        os.makedirs(os.path.dirname(p), exist_ok=True)
        with open(p, 'w') as f:
            f.write('# TEST TEXT')

    def discard(relative_path: str) -> None:
        os.unlink(os.path.join(root_ns_dir, relative_path))

    touchy('123.Qwerty.123.234.uavcan')
    touchy('nested/2.Asd.21.32.uavcan')
    touchy('nested/Foo.32.43.uavcan')

    dsdl_defs = _construct_dsdl_definitions_from_namespace(root_ns_dir)
    print(dsdl_defs)
    lut = {x.full_name: x for x in dsdl_defs}    # type: typing.Dict[str, _dsdl_definition.DSDLDefinition]
    assert len(lut) == 3

    assert str(lut['foo.Qwerty']) == repr(lut['foo.Qwerty'])
    assert str(lut['foo.Qwerty']) == \
        "DSDLDefinition(full_name='foo.Qwerty', version=Version(major=123, minor=234), fixed_port_id=123, " \
        "file_path=%r)" % lut['foo.Qwerty'].file_path

    assert str(lut['foo.nested.Foo']) == \
        "DSDLDefinition(full_name='foo.nested.Foo', version=Version(major=32, minor=43), fixed_port_id=None, " \
        "file_path=%r)" % lut['foo.nested.Foo'].file_path

    t = lut['foo.Qwerty']
    assert t.file_path == os.path.join(root_ns_dir, '123.Qwerty.123.234.uavcan')
    assert t.has_fixed_port_id
    assert t.fixed_port_id == 123
    assert t.text == '# TEST TEXT'
    assert t.version.major == 123
    assert t.version.minor == 234
    assert t.name_components == ['foo', 'Qwerty']
    assert t.short_name == 'Qwerty'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo'

    t = lut['foo.nested.Asd']
    assert t.file_path == os.path.join(root_ns_dir, 'nested', '2.Asd.21.32.uavcan')
    assert t.has_fixed_port_id
    assert t.fixed_port_id == 2
    assert t.text == '# TEST TEXT'
    assert t.version.major == 21
    assert t.version.minor == 32
    assert t.name_components == ['foo', 'nested', 'Asd']
    assert t.short_name == 'Asd'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo.nested'

    t = lut['foo.nested.Foo']
    assert t.file_path == os.path.join(root_ns_dir, 'nested', 'Foo.32.43.uavcan')
    assert not t.has_fixed_port_id
    assert t.fixed_port_id is None
    assert t.text == '# TEST TEXT'
    assert t.version.major == 32
    assert t.version.minor == 43
    assert t.name_components == ['foo', 'nested', 'Foo']
    assert t.short_name == 'Foo'
    assert t.root_namespace == 'foo'
    assert t.full_namespace == 'foo.nested'

    touchy('nested/Malformed.MAJOR.MINOR.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/Malformed.MAJOR.MINOR.uavcan')
    else:       # pragma: no cover
        assert False

    touchy('nested/NOT_A_NUMBER.Malformed.1.0.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/NOT_A_NUMBER.Malformed.1.0.uavcan')
    else:       # pragma: no cover
        assert False

    touchy('nested/Malformed.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
        discard('nested/Malformed.uavcan')
    else:       # pragma: no cover
        assert False

    _construct_dsdl_definitions_from_namespace(root_ns_dir)  # making sure all errors are cleared

    touchy('nested/super.bad/Unreachable.1.0.uavcan')
    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir)
    except FileNameFormatError as ex:
        print(ex)
    else:       # pragma: no cover
        assert False

    try:
        _construct_dsdl_definitions_from_namespace(root_ns_dir + '/nested/super.bad')
    except FileNameFormatError as ex:
        print(ex)
    else:       # pragma: no cover
        assert False

    discard('nested/super.bad/Unreachable.1.0.uavcan')
